#!/usr/bin/env python
#
# rendertexturestack.py - The RenderTextureStack class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`RenderTextureStack` class, which is used
by the :class:`.SliceCanvas` class to store a collection of off-screen
:class:`.RenderTexture` instances containing rendered slices of
:class:`.GLObject` instances.
"""


from __future__ import division

import logging

import numpy     as np
import OpenGL.GL as gl

import fsleyes.gl.routines as glroutines
import fsl.utils.idle      as idle
from . import                 rendertexture


log = logging.getLogger(__name__)


class RenderTextureStack(object):
    """The ``RenderTextureStack`` class creates and maintains a collection of
    :class:`.RenderTexture` instances, each of which is used to display a
    single slice of a :class:`.GLObject` along a specific display axis.

    The purpose of the ``RenderTextureStack`` is to pre-generate 2D slices of
    a :class:`.GLObject` so that they do not have to be rendered on-demand.
    Rendering a ``GLObject`` slices from a pre-generated off-screen texture
    provides better performance than rendering the ``GLObject`` slice
    in real time.

    The :class:`.RenderTexture` textures are updated in an idle loop, via the
    :func:`.idle.idle` function.


    .. note:: A ``RenderTextureStack`` instance must be manually updated
              whenever its ``GLObject`` changes, via the
              :meth:`onGLObjectUpdate` method. ``RenderTextureStack``
              instances do not explicitly listen for ``GLObject`` changes
              themselves, because there is no guarantee that the textures
              will be refreshed before they need to be drawn.
    """


    def __init__(self, globj):
        """Create a ``RenderTextureStack``. An update listener is registered
        on the ``GLObject``, so that the textures can be refreshed whenever it
        changes.

        :arg globj: The :class:`.GLObject` instance.
        """

        self.name = '{}_{}_{}'.format(
            type(self).__name__,
            type(globj).__name__, id(self))

        self.__globj              = globj
        self.__maxNumTextures     = 256
        self.__maxWidth           = 1024
        self.__maxHeight          = 1024
        self.__defaultNumTextures = 64
        self.__defaultWidth       = 256
        self.__defaultHeight      = 256

        self.__textureDirty       = []
        self.__textures           = []

        self.__lastDrawnTexture   = None
        self.__updateQueue        = []

        idle.idle(self.__textureUpdateLoop)

        log.debug('{}.init ({})'.format(type(self).__name__, id(self)))


    def __del__(self):
        """Prints a log message."""
        if log:
            log.debug('{}.del ({})'.format(type(self).__name__, id(self)))


    def destroy(self):
        """Must be called when this ``RenderTextureStack`` is no longer needed.
        Calls the :meth:`__destroyTextures` method.
        """
        self.__destroyTextures()


    def getGLObject(self):
        """Returns the :class:`.GLObject` associated with this
        ``RenderTextureStack``.
        """
        return self.__globj


    def draw(self, zpos, xform=None):
        """Draws the pre-generated :class:`.RenderTexture` which corresponds
        to the  specified Z position.

        :arg zpos:  Position of slice to render.

        :arg xform: Transformation matrix to apply to rendered slice vertices.
        """

        xax = self.__xax
        yax = self.__yax

        texIdx                  = self.__zposToIndex(zpos)
        self.__lastDrawnTexture = texIdx

        if texIdx < 0 or texIdx >= len(self.__textures):
            return

        lo, hi  = self.__globj.getDisplayBounds()
        texture = self.__textures[texIdx]

        if self.__textureDirty[texIdx]:
            self.__refreshTexture(texture, texIdx)

        log.debug('Drawing pre-rendered texture '
                  '[zax {}]: (zpos {}, slice {})'.format(
                      self.__zax,
                      zpos,
                      texIdx))

        texture.drawOnBounds(
            zpos, lo[xax], hi[xax], lo[yax], hi[yax], xax, yax, xform)


    def setAxes(self, xax, yax):
        """This method must be called when the display orientation of the
        :class:`.GLObject` changes. It destroys and re-creates all
        :class:`.RenderTexture` instances.
        """

        zax        = 3 - xax - yax
        self.__xax = xax
        self.__yax = yax
        self.__zax = zax

        res = self.__globj.getDataResolution(xax, yax)

        if res is not None: numTextures = res[zax]
        else:               numTextures = self.__defaultNumTextures

        if numTextures > self.__maxNumTextures:
            numTextures = self.__maxNumTextures

        self.__destroyTextures()

        for i in range(numTextures):
            rt = rendertexture.RenderTexture(
                '{}_{}'.format(self.name, i), rttype='c')
            self.__textures.append(rt)

        self.onGLObjectUpdate()


    def __destroyTextures(self):
        """Destroys all :class:`.RenderTexture` instances. This is performed
        asynchronously, via the ``idle.idle`` function.
        """

        texes = self.__textures
        self.__textures = []

        for tex in texes:
            idle.idle(tex.destroy)


    def onGLObjectUpdate(self):
        """Must be called called when the :class:`.GLObject` display is
        updated. Re-calculates the display space Z-axis range, and marks
        all render textures as dirty.
        """

        lo, hi      = self.__globj.getDisplayBounds()
        self.__zmin = lo[self.__zax]
        self.__zmax = hi[self.__zax]

        log.debug('GLObject range [zax {}]: {} - {}'.format(
            self.__zax,
            self.__zmin,
            self.__zmax))

        self.__refreshAllTextures()


    def __refreshAllTextures(self, *a):
        """Marks all :class:`.RenderTexture`  instances as *dirty*, so that
        they will be refreshed by the :meth:`.__textureUpdateLoop`.
        """

        if self.__lastDrawnTexture is not None:
            lastIdx = self.__lastDrawnTexture
        else:
            lastIdx = len(self.__textures) // 2

        aboveIdxs = list(range(lastIdx, len(self.__textures)))
        belowIdxs = list(range(lastIdx - 1, -1, -1))

        idxs = [0] * len(self.__textures)

        for i in range(len(self.__textures)):

            if len(aboveIdxs) > 0 and len(belowIdxs) > 0:
                if i % 2: idxs[i] = belowIdxs.pop(0)
                else:     idxs[i] = aboveIdxs.pop(0)

            elif len(aboveIdxs) > 0: idxs[i] = aboveIdxs.pop(0)
            else:                    idxs[i] = belowIdxs.pop(0)

        self.__textureDirty = [True] * len(self.__textures)
        self.__updateQueue  = idxs

        idle.idle(self.__textureUpdateLoop)


    def __textureUpdateLoop(self):
        """This method is called via the :func:`.idle.idle` function.
        It loops through all :class:`.RenderTexture` instances, and
        refreshes any that have been marked as *dirty*.

        Each call to this method causes one ``RenderTexture`` to be
        refreshed. After a ``RenderTexture`` has been refreshed, if there
        are dirty more ``RenderTexture`` instances, this method re-schedules
        itself to be called again via :func:`.idle.idle`.
        """

        if len(self.__updateQueue) == 0 or len(self.__textures) == 0:
            return

        idx = self.__updateQueue.pop(0)

        if self.__textureDirty[idx]:

            tex = self.__textures[idx]

            log.debug('Refreshing texture slice {} (zax {})'.format(
                idx, self.__zax))

            self.__refreshTexture(tex, idx)

        if len(self.__updateQueue) > 0:
            idle.idle(self.__textureUpdateLoop)


    def __refreshTexture(self, tex, idx):
        """Refreshes the given :class:`.RenderTexture`.

        :arg tex: The ``RenderTexture`` to refresh.
        :arg idx: Index of the ``RenderTexture``.
        """

        globj = self.__globj
        zpos  = self.__indexToZpos(idx)
        xax   = self.__xax
        yax   = self.__yax
        axes  = (self.__xax, self.__yax, self.__zax)

        if not globj.ready():
            return

        lo, hi = globj.getDisplayBounds()
        res    = globj.getDataResolution(xax, yax)

        if res is not None:
            width  = res[xax]
            height = res[yax]
        else:
            width  = self.__defaultWidth
            height = self.__defaultHeight

        if width  > self.__maxWidth:  width  = self.__maxWidth
        if height > self.__maxHeight: height = self.__maxHeight

        log.debug('Refreshing render texture for slice {} (zpos {}, '
                  'zax {}): {} x {}'.format(idx, zpos, self.__zax,
                                            width, height))

        tex.shape = width, height

        oldSize       = gl.glGetIntegerv(gl.GL_VIEWPORT)
        oldProjMat    = gl.glGetFloatv(  gl.GL_PROJECTION_MATRIX)
        oldMVMat      = gl.glGetFloatv(  gl.GL_MODELVIEW_MATRIX)

        with tex.target():
            glroutines.show2D(xax, yax, width, height, lo, hi)
            glroutines.clear((0, 0, 0, 0))

            with glroutines.disabled(gl.GL_BLEND):
                globj.preDraw()
                globj.draw2D(zpos, axes)
                globj.postDraw()

        gl.glViewport(*oldSize)
        gl.glMatrixMode(gl.GL_PROJECTION)
        gl.glLoadMatrixf(oldProjMat)
        gl.glMatrixMode(gl.GL_MODELVIEW)
        gl.glLoadMatrixf(oldMVMat)

        self.__textureDirty[idx] = False


    def __zposToIndex(self, zpos):
        """Converts a Z location in the display coordinate system into a
        ``RenderTexture`` index.
        """
        zmin  = self.__zmin
        zmax  = self.__zmax
        ntexs = len(self.__textures)
        step  = (zmax - zmin) / float(ntexs)

        # Round to avoid floating
        # point imprecision.
        index = np.around((zpos - zmin) / step, 5)

        # Be a little bit lenient at the boundariese
        if   abs(index)         < 0.01: index =  0
        elif abs(index - ntexs) < 0.01: index = ntexs - 1

        # But make sure that negative indices stay
        # negative, as otherwise, values between
        # -0.5 and 0 will be rounded to 0 in the
        # int(index) call below
        elif index < 0:                 index = -1

        return int(index)


    def __indexToZpos(self, index):
        """Converts a ``RenderTexture`` index into a Z location in the display
        coordinate system.
        """
        zmin  = self.__zmin
        zmax  = self.__zmax
        ntexs = len(self.__textures)
        step  = (zmax - zmin) / float(ntexs)

        return index * (zmax - zmin) / ntexs + (zmin + 0.5 * step)



#!/usr/bin/env python
#
# importdataseries.py - The ImportDataSeriesAction.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ImportDataSeriesAction`, which is used by
:class:`.PlotPanel` views to import data series from a text file.
"""

import            os
import os.path as op
import numpy   as np

import fsleyes_widgets.numberdialog as numdlg
import fsl.data.image               as fslimage
import fsl.utils.settings           as fslsettings

import fsleyes.strings              as strings
import fsleyes.plotting             as plotting
import fsleyes.colourmaps           as fslcm


from . import base


class ImportDataSeriesAction(base.Action):

    def __init__(self, overlayList, displayCtx, plotPanel):

        base.Action.__init__(self, overlayList, displayCtx, self.__doImport)
        self.__plotPanel = plotPanel


    def __doImport(self):

        import wx

        frame = wx.GetApp().GetTopWindow()

        # Ask the user where to get the data
        msg     = strings.messages[self, 'selectFile']
        fromDir = fslsettings.read('loadSaveOverlayDir', os.getcwd())
        dlg     = wx.FileDialog(frame,
                                message=msg,
                                defaultDir=fromDir,
                                style=wx.FD_OPEN)

        if dlg.ShowModal() != wx.ID_OK:
            return

        filePath = dlg.GetPath()
        fileName = op.basename(filePath)

        # Load the file, show an
        # error if it fails
        try:

            # Assuming that the data series
            # to plot are stored as columns
            data = np.loadtxt(filePath, dtype=float).T

            # Make sure the data is 2D, to
            # make code below easier and
            # happier.
            if len(data.shape) == 1:
                data = data.reshape((1, -1))

        except Exception as e:
            title = strings.titles[  self, 'error']
            msg   = strings.messages[self, 'error'].format(
                filePath,
                '{}: {}'.format(type(e).__name__, str(e)))

            wx.MessageBox(msg, title, wx.ICON_ERROR | wx.OK)
            return

        fslsettings.write('loadSaveOverlayDir', filePath)

        # Ask the user the x axis scaling factor.
        # If the currently selected overlay is
        # Nifti and 4D, default to its pixdim[3]
        overlay = self.displayCtx.getSelectedOverlay()

        if overlay is not None                 and \
           isinstance(overlay, fslimage.Nifti) and \
           len(overlay.shape) == 4             and \
           self.__plotPanel.usePixdim:
            xscale = overlay.pixdim[3]

        else:
            xscale = 1

        title = strings.titles[  self, 'selectXScale']
        msg   = strings.messages[self, 'selectXScale']

        # If the user pushes 'Ok', the entered value
        # is used as a fixed X axis interval. Otherwise,
        # it is assumed that the first column in the
        # file is the x axis data.
        dlg   = numdlg.NumberDialog(
            frame,
            title=title,
            message=msg,
            initial=xscale,
            minValue=1e-5,
            cancelText=strings.labels[self, 'firstColumnIsX'])

        firstColumnIsX = dlg.ShowModal() != wx.ID_OK
        xscale         = dlg.GetValue()

        # Add the data series
        series = []

        if firstColumnIsX:
            xdata = data[0,  :]
            ydata = data[1:, :]
        else:
            xdata = np.arange(0, data.shape[1] * xscale, xscale, dtype=float)
            ydata = data

        for i, ydata in enumerate(ydata):

            x   = np.array(xdata)
            y   = np.array(ydata)
            fin = np.isfinite(x) & np.isfinite(y)
            x   = x[fin]
            y   = y[fin]

            ds = plotting.DataSeries(None,
                                     self.overlayList,
                                     self.displayCtx,
                                     self.__plotPanel.canvas)
            ds.setData(x, y)

            # If we recognise the file name,
            # we can give it a useful label.
            label = strings.plotLabels.get(
                '{}.{}'  .format(fileName, i),
                '{} [{}]'.format(fileName, i))

            ds.label     = label
            ds.lineWidth = 1
            ds.colour    = fslcm.randomDarkColour()

            series.append(ds)

        self.__plotPanel.canvas.dataSeries.extend(series)



#!/usr/bin/env python
#
# wxglcolourbarcanvas.py - The WXGLColourBarCanvas class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`WXGLColourBarCanvas` class, which is a
:class:`.ColourBarCanvas` for use in a :mod:`wx` application.
"""


import six

import wx
import wx.glcanvas as wxgl

import fsleyes.gl                 as fslgl
import fsleyes.gl.colourbarcanvas as cbarcanvas


class WXGLColourBarCanvas(six.with_metaclass(fslgl.WXGLMetaClass,
                                             cbarcanvas.ColourBarCanvas,
                                             fslgl.WXGLCanvasTarget,
                                             wxgl.GLCanvas)):
    """The ``WXGLColourBarCanvas`` is a :class:`.ColourBarCanvas`, a
    :class:`wx.glcanvas.GLCanvas` and a :class:`.WXGLCanvasTarget`. If you
    want to use a :class:`.ColourBarCanvas` in your :mod:`wx` application, then
    you should use a ``WXGLColourBarCanvas``.

    .. note:: The ``WXGLColourBarCanvas`` assumes the existence of the
              :meth:`.ColourBarCanvas.updateColourBarTexture` method.
    """
    def __init__(self, parent, overlayList, displayCtx):

        attrs = fslgl.WXGLCanvasTarget.displayAttribues()

        wxgl.GLCanvas             .__init__(self, parent, **attrs)
        fslgl.WXGLCanvasTarget    .__init__(self)
        cbarcanvas.ColourBarCanvas.__init__(self, overlayList, displayCtx)

        def onsize(ev):
            self.updateColourBarTexture()
            ev.Skip()

        self.Bind(wx.EVT_SIZE, onsize)


    def destroy(self):
        """Must be called when this ``WXGLColourBarCanvas`` is no longer
        needed. Clears some event listeners and calls the base class
        ``destroy`` method.
        """
        self.Unbind(wx.EVT_SIZE)

        cbarcanvas.ColourBarCanvas.destroy(self)
        fslgl.WXGLCanvasTarget    .destroy(self)



#!/usr/bin/env python
#
# lightboxopts.py - The LightBoxOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LightBoxOpts` class, which is used
by :class:`.LightBoxPanel` instances for managing their display settings.
"""


import logging
import copy

from . import sceneopts
from . import canvasopts


log = logging.getLogger(__name__)


class LightBoxOpts(sceneopts.SceneOpts):
    """The ``LightBoxOpts`` class contains display settings for the
    :class:`.LightBoxPanel` class.

    All of the properties in the ``LightBoxOpts`` class are defined in the
    :class:`.LightBoxCanvasOpts` class - see its documentation for more
    details.
    """

    sliceSpacing   = copy.copy(canvasopts.LightBoxCanvasOpts.sliceSpacing)
    zax            = copy.copy(canvasopts.LightBoxCanvasOpts.zax)
    ncols          = copy.copy(canvasopts.LightBoxCanvasOpts.ncols)
    nrows          = copy.copy(canvasopts.LightBoxCanvasOpts.nrows)
    topRow         = copy.copy(canvasopts.LightBoxCanvasOpts.topRow)
    zrange         = copy.copy(canvasopts.LightBoxCanvasOpts.zrange)
    showGridLines  = copy.copy(canvasopts.LightBoxCanvasOpts.showGridLines)
    highlightSlice = copy.copy(canvasopts.LightBoxCanvasOpts.highlightSlice)


    def __init__(self, *args, **kwargs):
        """Create a ``LightBoxOpts`` instance. All arguments are passed
        through to the :class:`.SceneOpts` constructor.

        The :attr:`.SceneOpts.zoom` attribute is modified, as
        :class:`LightBoxPanel` uses it slightly differently to the
        :class:`OrthoPanel`.
        """
        sceneopts.SceneOpts.__init__(self, *args, **kwargs)
        self.zax = 2
        self.setAttribute('zax',  'default', 2)
        self.setAttribute('zoom', 'minval',  10)
        self.setAttribute('zoom', 'maxval',  1000)


    def _onPerformanceChange(self, *a):
        """Overrides :meth:`.SceneOpts._onPerformanceChange`. Changes the
        value of the :attr:`renderMode` and :attr:`highDpi` properties
        according to the performance setting.
        """

        if   self.performance == 3: self.renderMode = 'onscreen'
        elif self.performance == 2: self.renderMode = 'prerender'
        elif self.performance == 1: self.renderMode = 'prerender'

        self.highDpi = self.performance == 3 and self.highDpi

        log.debug('Performance settings changed: '
                  'renderMode={}'.format(self.renderMode))



#!/usr/bin/env python
#
# orthoopts.py - The OrthoOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoOpts` class, which contains display
settings used by the :class:`.OrthoPanel` class.
"""


import logging
import copy

import fsleyes_props      as props

from . import sceneopts
from . import canvasopts


log = logging.getLogger(__name__)


class OrthoOpts(sceneopts.SceneOpts):
    """The ``OrthoOpts`` class is used by :class:`.OrthoPanel` instances to
    manage their display settings.


    .. note:: While the ``OrthoOpts`` class has :attr:`xzoom`, :attr:`yzoom`,
              and :attr:`zzoom`, properties which control the zoom levels on
              each canvas independently, ``OrthoOpts`` class also inherits a
              ``zoom`` property from the :class:`.SceneOpts` class. This
              *global* zoom property can be used to adjust all canvas zoom
              levels simultaneously.
    """


    cursorGap = copy.copy(canvasopts.SliceCanvasOpts.cursorGap)


    showXCanvas = props.Boolean(default=True)
    """Toggles display of the X canvas."""


    showYCanvas = props.Boolean(default=True)
    """Toggles display of the Y canvas."""


    showZCanvas = props.Boolean(default=True)
    """Toggles display of the Z canvas."""


    showLabels = props.Boolean(default=True)
    """If ``True``, labels showing anatomical orientation are displayed on
    each of the canvases.
    """


    layout = props.Choice(('horizontal', 'vertical', 'grid'))
    """How should we lay out each of the three canvases?"""


    xzoom = copy.copy(sceneopts.SceneOpts.zoom)
    """Controls zoom on the X canvas."""


    yzoom = copy.copy(sceneopts.SceneOpts.zoom)
    """Controls zoom on the Y canvas."""


    zzoom = copy.copy(sceneopts.SceneOpts.zoom)
    """Controls zoom on the Z canvas. """


    invertXHorizontal = copy.copy(canvasopts.SliceCanvasOpts.invertX)
    """Invert the X canvas along the horizontal axis. """


    invertXVertical = copy.copy(canvasopts.SliceCanvasOpts.invertY)
    """Invert the X canvas along the vertical axis. """


    invertYHorizontal = copy.copy(canvasopts.SliceCanvasOpts.invertX)
    """Invert the Y canvas along the horizontal axis. """


    invertYVertical = copy.copy(canvasopts.SliceCanvasOpts.invertY)
    """Invert the Y canvas along the vertical axis. """


    invertZHorizontal = copy.copy(canvasopts.SliceCanvasOpts.invertX)
    """Invert the Z canvas along the horizontal axis. """


    invertZVertical = copy.copy(canvasopts.SliceCanvasOpts.invertY)
    """Invert the Z canvas along the vertical axis. """


    def __init__(self, *args, **kwargs):
        """Create an ``OrthoOpts`` instance. All arguments are passed
        through to the :class:`.SceneOpts` constructor.

        This method sets up a binding from the :attr:`.SceneOpts.zoom`
        property to the :attr:`xzoom`, :attr:`yzoom`, and :attr:`zzoom`
        properties - see :meth:`__onZoom`.
        """
        sceneopts.SceneOpts.__init__(self, *args, **kwargs)

        name = '{}_{}'.format(type(self).__name__, id(self))

        self.addListener('zoom', name, self.__onZoom)


    def __onZoom(self, *a):
        """Called when the :attr:`.SceneOpts.zoom` property changes.

        Propagates the change to the :attr:`xzoom`, :attr:`yzoom`, and
        :attr:`zzoom` properties.
        """
        self.xzoom = self.zoom
        self.yzoom = self.zoom
        self.zzoom = self.zoom


    def _onPerformanceChange(self, *a):
        """Overrides :meth:`.SceneOpts._onPerformanceChange`. Changes the
        value of the :attr:`renderMode` and :attr:`highDpi` properties
        according to the performance setting.
        """

        if   self.performance == 3: self.renderMode = 'onscreen'
        elif self.performance == 2: self.renderMode = 'offscreen'
        elif self.performance == 1: self.renderMode = 'prerender'

        self.highDpi = self.performance == 3 and self.highDpi

        log.debug('Performance settings changed: '
                  'renderMode={}'.format(self.renderMode))



#!/usr/bin/env python
#
# layout.py - The layout API (previously called "perspectives").
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides functions for managing *layouts* - stored view and
control panel layouts for *FSLeyes*. Layouts may be persisted using the
:mod:`.settings` module. A few layouts are also *built in*, and are
defined in the :attr:`BUILT_IN_LAYOUTS` dictionary.


.. note:: Prior to FSLeyes 0.24.0, *layouts* were called *perspectives*.


The ``layouts`` module provides the following functions. These are intended
for use by the :class:`.FSLeyesFrame`, but can be used in other ways too:


.. autosummary::
   :nosignatures:

   getAllLayouts
   loadLayout
   applyLayout
   saveLayout
   removeLayout
   serialiseLayout
   deserialiseLayout


A layout defines a layout for a :class:`.FSLeyesFrame`. It specifies the
type and layout of one or more *views* (defined in the :mod:`.views` module)
and, within each view, the type and layout of one or more *controls* (defined
in the :mod:`.controls` module). See the :mod:`fsleyes` documentation for
an overview of views and controls.


All of this information is stored as a string - see the
:func:`serialiseLayout` function for details on its storage format.
"""


import functools as ft
import              logging
import              pkgutil
import              textwrap
import              importlib
import              collections

import fsl.utils.settings            as fslsettings
import fsleyes_widgets.utils.status  as status
import fsleyes.strings               as strings
import fsleyes.plugins               as plugins
import fsleyes.controls              as controls
import fsleyes.views                 as views
import fsleyes.views.viewpanel       as viewpanel
import fsleyes.views.canvaspanel     as canvaspanel
import fsleyes.views.plotpanel       as plotpanel
import fsleyes.controls.controlpanel as controlpanel


log = logging.getLogger(__name__)


def getAllLayouts():
    """Returns a list containing the names of all saved layouts. The
    returned list does not include built-in layouts - these are
    accessible in the :attr:`BUILT_IN_LAYOUTS` dictionary.
    """

    layouts = fslsettings.read('fsleyes.layouts',      []) + \
              fslsettings.read('fsleyes.perspectives', [])

    uniq = []
    for l in layouts:
        if l not in uniq:
            uniq.append(l)

    return uniq


def loadLayout(frame, name, **kwargs):
    """Load the named layout, and apply it to the given
    :class:`.FSLeyesFrame`. The ``kwargs`` are passed through to the
    :func:`applyLayout` function.
    """

    if name in BUILT_IN_LAYOUTS.keys():

        log.debug('Loading built-in layout {}'.format(name))
        layout = BUILT_IN_LAYOUTS[name]

    else:
        log.debug('Loading saved layout {}'.format(name))
        layout = fslsettings.read('fsleyes.layouts.{}'.format(name), None)
        if layout is None:
            fslsettings.read('fsleyes.perspectives.{}'.format(name), None)

    if layout is None:
        raise ValueError('No layout named "{}" exists'.format(name))

    log.debug('Applying layout:\n{}'.format(layout))
    applyLayout(frame, name, layout, **kwargs)


def applyLayout(frame, name, layout, message=None):
    """Applies the given serialised layout string to the given
    :class:`.FSLeyesFrame`.

    :arg frame:   The :class:`.FSLeyesFrame` instance.
    :arg name:    The layout name.
    :arg layout:  The serialised layout string.
    :arg message: A message to display (using the :mod:`.status` module).
    """

    import fsleyes.views.canvaspanel as canvaspanel

    layout        = deserialiseLayout(layout)
    frameChildren = layout[0]
    frameLayout   = layout[1]
    vpChildrens   = layout[2]
    vpLayouts     = layout[3]
    vpPanelProps  = layout[4]
    vpSceneProps  = layout[5]

    # Show a message while re-configuring the frame
    if message is None:
        message = strings.messages[
            'layout.applyingLayout'].format(
                strings.layouts.get(name, name))

    status.update(message)

    # Clear all existing view
    # panels from the frame
    frame.removeAllViewPanels()

    # Add all of the view panels
    # specified in the layout
    for vp in frameChildren:
        log.debug('Adding view panel {} to frame'.format(vp.__name__))
        frame.addViewPanel(vp, defaultLayout=False)

    # Apply the layout to those view panels
    frame.auiManager.LoadPerspective(frameLayout)

    # For each view panel, add all of the
    # control panels, and lay them out
    viewPanels = frame.viewPanels
    for i in range(len(viewPanels)):

        vp         = viewPanels[  i]
        children   = vpChildrens[ i]
        vpLayout   = vpLayouts[   i]
        panelProps = vpPanelProps[i]
        sceneProps = vpSceneProps[i]

        for child in children:
            log.debug('Adding control panel {} to {}'.format(
                child.__name__, type(vp).__name__))
            _addControlPanel(vp, child)

        vp.auiManager.LoadPerspective(vpLayout)

        # Apply saved property values
        # to the view panel.
        for name, val in panelProps.items():
            log.debug('Setting {}.{} = {}'.format(
                type(vp).__name__, name, val))
            vp.deserialise(name, val)

        # And to its SceneOpts instance if
        # it is a CanvasPanel, or its
        # PlotCanvas if it is a PlotPanel
        if   isinstance(vp, canvaspanel.CanvasPanel): aux = vp.sceneOpts
        elif isinstance(vp, plotpanel.PlotPanel):     aux = vp.canvas

        for name, val in sceneProps.items():
            log.debug('Setting {}.{} = {}'.format(
                type(aux).__name__, name, val))
            aux.deserialise(name, val)


def saveLayout(frame, name):
    """Serialises the layout of the given :class:`.FSLeyesFrame` and saves
    it as a layout with the given name.
    """

    if name in BUILT_IN_LAYOUTS.keys():
        raise ValueError('A built-in layout named "{}" '
                         'already exists'.format(name))

    log.debug('Saving current layout with name {}'.format(name))

    layout = serialiseLayout(frame)
    fslsettings.write('fsleyes.layouts.{}'.format(name), layout)

    _addToLayoutList(name)

    log.debug('Serialised layout:\n{}'.format(layout))


def removeLayout(name):
    """Deletes the named layout. """

    log.debug('Deleting layout with name {}'.format(name))
    fslsettings.delete('fsleyes.layouts.{}'     .format(name))
    fslsettings.delete('fsleyes.perspectives.{}'.format(name))
    _removeFromLayoutList(name)


def serialiseLayout(frame):
    """Serialises the layout of the given :class:`.FSLeyesFrame`, and returns
    it as a string.

    .. note:: This function was written against wx.lib.agw.aui.AuiManager as
              it exists in wxPython 3.0.2.0.

     *FSLeyes* uses a hierarchy of ``wx.lib.agw.aui.AuiManager`` instances for
     its layout - the :class:`.FSLeyesFrame` uses an ``AuiManager`` to lay out
     :class:`.ViewPanel` instances, and each of these ``ViewPanels`` use their
     own ``AuiManager`` to lay out control panels.

     The layout for a single ``AuiManager`` can be serialised to a string via
     the ``AuiManager.SavePerspective`` and ``AuiManager.SavePaneInfo``
     methods. One of these strings consists of:

       - A name, `'layout1'` or `'layout2'`, specifying the AUI version
         (this will always be at least `'layout2'` for *FSLeyes*).

       - A set of key-value set of key-value pairs defining the top level
         panel layout.

       - A set of key-value pairs for each pane, defining its layout. the
         ``AuiManager.SavePaneInfo`` method returns this for a single pane.

     These are all encoded in a single string, with the above components
     separated with '|' characters, and the pane-level key-value pairs
     separated with a ';' character. For example:

     layout2|key1=value1|name=Pane1;caption=Pane 1|\
     name=Pane2;caption=Pane 2|doc_size(5,0,0)=22|

     This function queries each of the AuiManagers, and extracts the following:

        1. A layout string for the :class:`.FSLeyesFrame`.

        2. A string containing a comma-separated list of :class:`.ViewPanel`
           class names, in the same order as they are specified in the frame
           layout string.

        3.  For each ``ViewPanel``:

             - A layout string for the ``ViewPanel``

             - A string containing a comma-separated list of control panel
               class names, in the same order as specified in the
               ``ViewPanel`` layout string.

    Each of these pieces of information are then concatenated into a single
    newline separated string.

    In FSLeyes 0.35.0, the list of ``ViewPanel`` and ``ControlPanel`` class
    names was changed from containing just the class names
    (e.g. ``'OrthoPanel'``) to containing the fully resolved class paths
    (e.g. ``'fsleyes.views.orthopanel.OrthoPanel'``). The
    :func:`deserialiseLayout` function is compatible with both formats.
    """

    # We'll start by defining this silly function, which
    # takes an ``AuiManager`` layout string, and a list
    # of the children which are being managed by the
    # AuiManager, and makes sure that the order of the
    # child pane layout specifications in the string is
    # the same as the order of the children in the list.
    #
    # If the 'rename' argument is True, this function
    # performs an additional step.
    #
    # The FSLeyesFrame gives each of its view panels a
    # unique name of the form "ClassName index", where
    # the 'index' is a sequentially increasing identifier
    # number (so that multiple views of the same type can
    # be differentiated). If the 'rename' argument to
    # this function is True, these names are adjusted so
    # that they begin at 1 and increase sequentially. This
    # is done by the patchPanelName function, defined
    # below.
    #
    # This name adjustment is required to handle
    # situations where the indices of existing view panels
    # are not sequential, as when a layout is applied, the
    # view panel names given by the FSLeyesFrame must
    # match the names that are specified in the layout
    # string.
    #
    # In addition to patching the name of each panel,
    # the 'rename' argument will also cause the panel
    # caption (its display title) to be adjusted so that
    # it is in line with the name.
    def patchLayoutString(auiMgr, panels, rename=False):

        layoutStr = auiMgr.SavePerspective()

        # The different sections of the string
        # returned by SavePerspective are
        # separated with a '|' character.
        sections = layoutStr.split('|')
        sections = [s.strip() for s in sections]
        sections = [s for s in sections if s != '']

        # Here, we identify sections which specify
        # the layout of a child pane, remove them,
        # and patch them back in, in the order that
        # the child panels are specified in the list.
        pi = 0
        for si, s in enumerate(sections):
            if s.find('name=') > -1:
                panel        = panels[pi]
                panelInfo    = auiMgr.GetPane(panel)
                panelLayout  = auiMgr.SavePaneInfo(panelInfo)
                pi          += 1
                sections[si] = panelLayout

                if rename:
                    sections[si] = patchPanelName(sections[si], pi)

        # Now the panel layouts in our layout string
        # are in the same order as our list of view
        # panels - we can re-join the layout string
        # sections, and we're done.
        return '|'.join(sections) + '|'

    # The purpose of this function is described above.
    def patchPanelName(layoutString, index):
        # In each AUI layout section, 'key=value'
        # pairs are separated with a semi-colon
        kvps = layoutString.split(';')

        # And each 'key=value' pair is separated
        # with an equals character
        kvps = [kvp.split('=') for kvp in kvps]
        kvps = collections.OrderedDict(kvps)

        # We need to update the indices contained
        # in the 'name' and 'caption' values
        name    = kvps['name']
        caption = kvps['caption']

        # Strip off the old index
        name    = ' '.join(name   .split()[:-1])
        caption = ' '.join(caption.split()[:-1])

        # Patch in the new index
        name    = '{} {}'.format(name,    index)
        caption = '{} {}'.format(caption, index)

        kvps['name']    = name
        kvps['caption'] = caption

        # Reconstruct the layout string
        kvps = ['='.join((k, v)) for k, v in kvps.items()]
        kvps = ';'.join(kvps)

        return kvps

    # Now we can start extracting the layout information.
    # We start with the FSLeyesFrame layout.
    auiMgr     = frame.auiManager
    viewPanels = frame.viewPanels

    # Generate the frame layout string, and a
    # list of the children of the frame
    frameLayout   = patchLayoutString(auiMgr, viewPanels, True)
    frameChildren = ['.'.join((type(vp).__module__, type(vp).__qualname__))
                     for vp in viewPanels]
    frameChildren = ','.join(frameChildren)

    # We are going to build a list of layout strings,
    # one for each ViewPanel, and a corresponding list
    # of control panels displayed on each ViewPanel.
    vpLayouts = []
    vpConfigs = []

    for vp in viewPanels:

        # Get the auiManager and layout for this view panel.
        # This is a little bit complicated, as ViewPanels
        # differentiate between the main 'centre' panel, and
        # all other secondary (control) panels. The layout
        # string needs to contain layout information for
        # all of these panels, but we only care about the
        # control panels.
        vpAuiMgr    = vp.auiManager
        ctrlPanels  = vp.getPanels()
        centrePanel = vp.centrePanel

        # As above for the frame, generate a layout
        # string and a list of control panels - the
        # children of the view panel.
        vpLayout    = patchLayoutString(vpAuiMgr, [centrePanel] + ctrlPanels)
        vpChildren  = ['.'.join((type(cp).__module__, type(cp).__qualname__))
                       for cp in ctrlPanels]
        vpChildren  = ','.join(vpChildren)

        # Get the panel and scene settings
        panelProps, sceneProps = _getPanelProps(vp)

        # And turn them into comma-separated key-value pairs.
        panelProps = ['{}={}'.format(k, v) for k, v in panelProps.items()]
        sceneProps = ['{}={}'.format(k, v) for k, v in sceneProps.items()]

        panelProps = ','.join(panelProps)
        sceneProps = ','.join(sceneProps)

        # Build the config string - the children,
        # the panel settings and the scene settings.
        vpConfig = ';'.join([vpChildren, panelProps, sceneProps])

        vpLayouts.append(vpLayout)
        vpConfigs.append(vpConfig)

    # We serialise all of these pieces of information
    # as a single newline-separated string.
    layout = [frameChildren, frameLayout]
    for vpConfig, vpLayout in zip(vpConfigs, vpLayouts):
        layout.append(vpConfig)
        layout.append(vpLayout)

    # And we're done!
    return '\n'.join(layout)


def deserialiseLayout(layout):
    """Deserialises a layout string which was created by the
    :func:`serialiseLayout` string.

    :returns: A tuple containing the following:

                - A list of :class:`.ViewPanel` class types - the
                  children of the :class:`.FSLeyesFrame`.

                - An ``aui`` layout string for the :class:`.FSLeyesFrame`

                - A list of lists, one for each ``ViewPanel``, with each
                  list containing a collection of control panel class
                  types - the children of the corresponding ``ViewPanel``.

                - A list of strings, one ``aui`` layout string for each
                  ``ViewPanel``.

                - A list of dictionaries, one for each ``ViewPanel``,
                  containing property ``{name : value}`` pairs to be
                  applied to the ``ViewPanel``.

                - A list of dictionaries, one for each ``ViewPanel``,
                  containing property ``{name : value}`` pairs to be applied
                  to the :class:`.SceneOpts` instance associated with the
                  ``ViewPanel``, if it is a :class:`.CanvasPanel`, or the
                  :class:`.PlotCanvas` instance associated with the
                  ``ViewPanel``, if it is a :class:`.PlotPanel`.
    """

    # Versions of FSLeyes prior to 1.0.0 would just
    # save the view/control class name. This was
    # changed in 1.0.0 so that the full path to the
    # class is saved. This function aims to be
    # compatible with both formats - given a class
    # name, or a fully resolved class name, it will
    # return the corresponding type object.
    def findViewOrControl(panelname, paneltype):

        # new format
        if '.' in panelname:
            mod, cls = panelname.rsplit('.', maxsplit=1)
            mod      = importlib.import_module(mod)
            return getattr(mod, cls)

        # make a list of all candidate types,
        # then search through them for a match
        panels = []

        # builtins
        if paneltype == 'control':
            basemod  = controls
            basetype = (controlpanel.ControlPanel,
                        controlpanel.ControlToolBar)
        else:
            basemod  = views
            basetype = viewpanel.ViewPanel

        mods = pkgutil.iter_modules(basemod.__path__, basemod.__name__ + '.')
        for _, mod, _  in mods:
            mod = importlib.import_module(mod)
            for att in dir(mod):
                att = getattr(mod, att)
                if isinstance(att, type) and issubclass(att, basetype):
                    panels.append(att)

        # plugins
        if paneltype == 'control':
            panels.extend(plugins.listControls().values())
        else:
            panels.extend(plugins.listViews().values())

        for panel in panels:
            if panel.__name__ == panelname:
                return panel

        raise ValueError('Unknown FSLeyes panel type: {}'.format(panelname))

    findView    = ft.partial(findViewOrControl, paneltype='view')
    findControl = ft.partial(findViewOrControl, paneltype='control')

    lines = layout.split('\n')
    lines = [line.strip() for line in lines]
    lines = [line         for line in lines if line != '']

    frameChildren = lines[0]
    frameLayout   = lines[1]

    # The children strings are comma-separated
    # class names. The frame children are ViewPanels,
    # which are all defined in the fsleyes.views
    # package.
    frameChildren = frameChildren.split(',')
    frameChildren = [fc.strip()   for fc in frameChildren]
    frameChildren = [fc           for fc in frameChildren if fc != '']
    frameChildren = [findView(fc) for fc in frameChildren]

    # Collate the children/layouts for each view panel
    vpChildren   = []
    vpLayouts    = []
    vpPanelProps = []
    vpSceneProps = []

    for i in range(len(frameChildren)):

        linei = (i * 2) + 2

        config = lines[linei]
        layout = lines[linei + 1]

        children, panelProps, sceneProps = config.split(';')

        vpChildren   .append(children)
        vpLayouts    .append(layout)
        vpPanelProps .append(panelProps)
        vpSceneProps .append(sceneProps)

    # The ViewPanel children string is a comma-separated
    # list of control panel class names. All control panels
    # should be defined in the fsleyes.controls package.
    for i in range(len(vpChildren)):

        children      = vpChildren[i].split(',')
        children      = [vpc.strip()      for vpc in children]
        children      = [vpc              for vpc in children if vpc != '']
        children      = [findControl(vpc) for vpc in children]
        vpChildren[i] = children

    # The panel props and scene props strings are
    # comma-separated lists of 'prop=value' pairs.
    # We'll turn them into a dict for convenience.
    for i in range(len(vpPanelProps)):
        props           = vpPanelProps[i].split(',')
        props           = [p for p in props if p != '']
        props           = [p.split('=') for p in props]
        vpPanelProps[i] = collections.OrderedDict(props)

    for i in range(len(vpSceneProps)):
        props           = vpSceneProps[i].split(',')
        props           = [p for p in props if p != '']
        props           = [p.split('=') for p in props]
        vpSceneProps[i] = collections.OrderedDict(props)

    return (frameChildren,
            frameLayout,
            vpChildren,
            vpLayouts,
            vpPanelProps,
            vpSceneProps)


def _addToLayoutList(layout):
    """Adds the given layout name to the list of saved layouts. """

    layout  = layout.strip()
    layouts = getAllLayouts()

    if layout not in layouts:
        layouts.append(layout)

    log.debug('Updating stored layout list: {}'.format(layout))
    fslsettings.write('fsleyes.layouts', layouts)


def _removeFromLayoutList(layout):
    """Removes the given layout name from the list of saved layouts.
    """

    layouts = getAllLayouts()

    try:               layouts.remove(layout)
    except ValueError: return

    log.debug('Updating stored layout list: {}'.format(layouts))
    fslsettings.write('fsleyes.layouts', layouts)


def _addControlPanel(viewPanel, panelType):
    """Adds a control panel to the given :class:`.ViewPanel`.

    :arg viewPanel: A :class:`.ViewPanel` instance.
    :arg panelType: A control panel type.
    """
    viewPanel.togglePanel(panelType)


def _getPanelProps(panel):
    """Creates and returns two dictionaries, containing properties of the given
    :class:`.ViewPanel` (and its associated :class:`.SceneOpts` instance, if
    it is a :class:`.CanvasPanel`, or :class:`.PlotCanvas`, if it is a
    :class:`.PlotPanel`), which are to be saved as part of a seriaised
    *FSLeyes* layout. The properties to be saved are listed in the
    :data:`VIEWPANEL_PROPS` dictionary.
    """

    if not isinstance(panel, (canvaspanel.CanvasPanel, plotpanel.PlotPanel)):
        return {}, {}

    panelType              = type(panel).__name__
    panelProps, sceneProps = VIEWPANEL_PROPS[panelType]

    if isinstance(panel, canvaspanel.CanvasPanel):
        aux = panel.sceneOpts
    elif isinstance(panel, plotpanel.PlotPanel):
        aux = panel.canvas

    panelProps = {name : panel.serialise(name) for name in panelProps}
    sceneProps = {name : aux  .serialise(name) for name in sceneProps}

    return panelProps, sceneProps


VIEWPANEL_PROPS = {
    'OrthoPanel'         : [['syncLocation',
                             'syncOverlayOrder',
                             'syncOverlayDisplay',
                             'syncOverlayVolume',
                             'movieRate',
                             'movieAxis'],
                            ['showCursor',
                             'bgColour',
                             'fgColour',
                             'cursorColour',
                             'cursorGap',
                             'showColourBar',
                             'colourBarLocation',
                             'colourBarLabelSide',
                             'showXCanvas',
                             'showYCanvas',
                             'showZCanvas',
                             'showLabels',
                             'labelSize',
                             'layout',
                             'xzoom',
                             'yzoom',
                             'zzoom',
                             'highDpi']],
    'LightBoxPanel'      : [['syncLocation',
                             'syncOverlayOrder',
                             'syncOverlayDisplay',
                             'syncOverlayVolume',
                             'movieRate',
                             'movieAxis'],
                            ['showCursor',
                             'bgColour',
                             'fgColour',
                             'cursorColour',
                             'showColourBar',
                             'colourBarLocation',
                             'colourBarLabelSide',
                             'zax',
                             'showGridLines',
                             'highlightSlice',
                             'highDpi']],
    'Scene3DPanel'       : [['syncLocation',
                             'syncOverlayOrder',
                             'syncOverlayDisplay',
                             'syncOverlayVolume'],
                            ['showCursor',
                             'bgColour',
                             'fgColour',
                             'cursorColour',
                             'showColourBar',
                             'colourBarLocation',
                             'colourBarLabelSide',
                             'occlusion',
                             'light',
                             'lightPos',
                             'offset',
                             'rotation',
                             'showLegend']],
    'TimeSeriesPanel'    : [['usePixdim',
                             'plotMode',
                             'plotMelodicICs'],
                            ['legend',
                             'xAutoScale',
                             'yAutoScale',
                             'xLogScale',
                             'yLogScale',
                             'ticks',
                             'grid',
                             'gridColour',
                             'bgColour',
                             'smooth']],
    'HistogramPanel'     : [['histType',
                             'plotType'],
                            ['legend',
                             'xAutoScale',
                             'yAutoScale',
                             'xLogScale',
                             'yLogScale',
                             'ticks',
                             'grid',
                             'gridColour',
                             'bgColour',
                             'smooth']],
    'PowerSpectrumPanel' : [['plotMelodicICs',
                             'plotFrequencies'],
                            ['legend',
                             'xAutoScale',
                             'yAutoScale',
                             'xLogScale',
                             'yLogScale',
                             'ticks',
                             'grid',
                             'gridColour',
                             'bgColour',
                             'smooth']]}


# The order in which properties are defined in
# a layout is the order in which they will
# be applied. This is important to remember when
# considering properties that have side effects
# (e.g. setting SceneOpts.bgColour will clobber
# SceneOpts.fgColour).
BUILT_IN_LAYOUTS = collections.OrderedDict((
    ('default',
     textwrap.dedent("""
                     fsleyes.views.orthopanel.OrthoPanel
                     layout2|name=OrthoPanel 1;caption=Ortho View 1;state=67376064;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=22|
                     fsleyes.controls.orthotoolbar.OrthoToolBar,fsleyes.controls.overlaydisplaytoolbar.OverlayDisplayToolBar,fsleyes.controls.overlaylistpanel.OverlayListPanel,fsleyes.controls.locationpanel.LocationPanel;syncOverlayOrder=True,syncLocation=True,syncOverlayDisplay=True,movieRate=400;colourBarLocation=top,showCursor=True,bgColour=#000000ff,layout=horizontal,colourBarLabelSide=top-left,cursorGap=False,fgColour=#ffffffff,cursorColour=#00ff00ff,showXCanvas=True,showYCanvas=True,showColourBar=False,showZCanvas=True,showLabels=True
                     layout2|name=Panel;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OrthoToolBar;caption=Ortho view toolbar;state=67382012;dir=1;layer=10;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OverlayDisplayToolBar;caption=Display toolbar;state=67382012;dir=1;layer=11;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OverlayListPanel;caption=Overlay list;state=67373052;dir=3;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=1;minh=1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=LocationPanel;caption=Location;state=67373052;dir=3;layer=0;row=0;pos=1;prop=100000;bestw=-1;besth=-1;minw=1;minh=1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=22|dock_size(3,0,0)=176|dock_size(1,10,0)=49|dock_size(1,11,0)=67|
                     """)),  # noqa

    ('melodic',
     textwrap.dedent("""
                     fsleyes.views.lightboxpanel.LightBoxPanel,fsleyes.views.timeseriespanel.TimeSeriesPanel,fsleyes.views.powerspectrumpanel.PowerSpectrumPanel
                     layout2|name=LightBoxPanel 1;caption=Lightbox View 1;state=67377088;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=TimeSeriesPanel 2;caption=Time series 2;state=67377148;dir=3;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=PowerSpectrumPanel 3;caption=Power spectra 3;state=67377148;dir=3;layer=0;row=0;pos=1;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=22|dock_size(3,0,0)=224|
                     fsleyes.controls.locationpanel.LocationPanel,fsleyes.controls.overlaylistpanel.OverlayListPanel,fsleyes.plugins.controls.melodicclassificationpanel.MelodicClassificationPanel,fsleyes.controls.lightboxtoolbar.LightBoxToolBar,fsleyes.controls.overlaydisplaytoolbar.OverlayDisplayToolBar;syncLocation=True,syncOverlayOrder=True,movieRate=750,syncOverlayDisplay=True;bgColour=#000000ff,fgColour=#ffffffff,showCursor=True,cursorColour=#00ff00ff,highlightSlice=False,zax=2,showColourBar=False,showGridLines=False,colourBarLocation=top
                     layout2|name=Panel;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=LocationPanel;caption=Location;state=67373052;dir=3;layer=0;row=0;pos=1;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OverlayListPanel;caption=Overlay list;state=67373052;dir=3;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=MelodicClassificationPanel;caption=Melodic IC classification;state=67373052;dir=2;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=LightBoxToolBar;caption=Lightbox view toolbar;state=67382012;dir=1;layer=10;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OverlayDisplayToolBar;caption=Display toolbar;state=67382012;dir=1;layer=11;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=22|dock_size(3,0,0)=130|dock_size(1,10,0)=45|dock_size(1,11,0)=51|dock_size(2,0,0)=402|
                     TimeSeriesToolBar;;
                     layout2|name=FigureCanvasWxAgg;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=TimeSeriesToolBar;caption=Time series toolbar;state=67382012;dir=1;layer=10;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=642|dock_size(1,10,0)=36|
                     PowerSpectrumToolBar;;
                     layout2|name=FigureCanvasWxAgg;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=PowerSpectrumToolBar;caption=Plot toolbar;state=67382012;dir=1;layer=10;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=642|dock_size(1,10,0)=36|
                     """)),  # noqa

    ('feat',
     textwrap.dedent("""
                     fsleyes.views.orthopanel.OrthoPanel,fsleyes.views.timeseriespanel.TimeSeriesPanel
                     layout2|name=OrthoPanel 1;caption=Ortho View 1;state=67377088;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=TimeSeriesPanel 2;caption=Time series 2;state=67377148;dir=3;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=22|dock_size(3,0,0)=282|
                     fsleyes.controls.overlaylistpanel.OverlayListPanel,fsleyes.controls.overlaydisplaytoolbar.OverlayDisplayToolBar,fsleyes.controls.orthotoolbar.OrthoToolBar,fsleyes.controls.locationpanel.LocationPanel,fsleyes.plugins.controls.clusterpanel.ClusterPanel;syncLocation=True,syncOverlayOrder=True,movieRate=750,syncOverlayDisplay=True;layout=horizontal,showLabels=True,bgColour=#000000ff,fgColour=#ffffffff,showCursor=True,showZCanvas=True,cursorColour=#00ff00ff,showColourBar=False,showYCanvas=True,showXCanvas=True,colourBarLocation=top
                     layout2|name=Panel;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OverlayListPanel;caption=Overlay list;state=67373052;dir=3;layer=2;row=0;pos=0;prop=87792;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OverlayDisplayToolBar;caption=Display toolbar;state=67382012;dir=1;layer=10;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OrthoToolBar;caption=Ortho view toolbar;state=67382012;dir=1;layer=10;row=1;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=LocationPanel;caption=Location;state=67373052;dir=3;layer=2;row=0;pos=1;prop=98544;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=ClusterPanel;caption=Cluster browser;state=67373052;dir=2;layer=1;row=0;pos=0;prop=114760;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=10|dock_size(2,1,0)=566|dock_size(1,10,0)=51|dock_size(1,10,1)=36|dock_size(3,2,0)=130|
                     OverlayListPanel,TimeSeriesToolBar;;
                     layout2|name=FigureCanvasWxAgg;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=OverlayListPanel;caption=Overlay list;state=67373052;dir=4;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|name=TimeSeriesToolBar;caption=Time series toolbar;state=67382012;dir=1;layer=10;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=642|dock_size(1,10,0)=36|dock_size(4,0,0)=206|
                     """)),  # noqa

    ('ortho',
     textwrap.dedent("""
                     fsleyes.views.orthopanel.OrthoPanel
                     layout2|name=OrthoPanel 1;caption=Ortho View 1;state=67376064;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=22|
                     ;syncLocation=True,syncOverlayOrder=True,syncOverlayDisplay=True;layout=horizontal,showLabels=True,bgColour=#000000ff,fgColour=#ffffffff,showCursor=True,showZCanvas=True,cursorColour=#00ff00ff,showColourBar=False,showYCanvas=True,showXCanvas=True,colourBarLocation=top
                     layout2|name=Panel;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=22|
                     """)),  # noqa
    ('3d',
     textwrap.dedent("""
                     fsleyes.views.scene3dpanel.Scene3DPanel
                     layout2|name=Scene3DPanel 1;caption=3D View 1;state=67376064;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=24|
                     ;syncOverlayOrder=True,syncOverlayDisplay=True,syncLocation=True;showColourBar=False,showLegend=True,cursorColour=#00ff00ff,colourBarLocation=top,showCursor=True,colourBarLabelSide=top-left,bgColour=#9999c0ff,fgColour=#00ff00ff
                     layout2|name=Panel;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=22|
                     """)),  # noqa

    ('lightbox',
     textwrap.dedent("""
                     fsleyes.views.lightboxpanel.LightBoxPanel
                     layout2|name=LightBoxPanel 1;caption=Lightbox View 1;state=67376064;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=22|
                     ;syncLocation=True,syncOverlayOrder=True,syncOverlayDisplay=True;bgColour=#000000ff,fgColour=#ffffffff,showCursor=True,cursorColour=#00ff00ff,highlightSlice=False,zax=2,showColourBar=False,showGridLines=False,colourBarLocation=top
                     layout2|name=Panel;caption=;state=768;dir=5;layer=0;row=0;pos=0;prop=100000;bestw=-1;besth=-1;minw=-1;minh=-1;maxw=-1;maxh=-1;floatx=-1;floaty=-1;floatw=-1;floath=-1;notebookid=-1;transparent=255|dock_size(5,0,0)=10|
                     """))))  # noqa



#!/usr/bin/env python
#
# histogramtoolbar.py - The HistogramToolBar class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`HistogramToolBar`, a toolbar for use with
a :class:`.HistogramPanel`.
"""


import fsleyes_props                as props
import fsleyes.icons                as icons
import fsleyes.strings              as strings
import fsleyes.tooltips             as tooltips
import fsleyes.actions              as actions
import fsleyes.views.histogrampanel as histogrampanel

from . import plottoolbar


class HistogramToolBar(plottoolbar.PlotToolBar):
    """The ``HistogramToolBar`` is a toolbar for use with a
    :class:`.HistogramPanel`. It extends :class:`.PlotToolBar`,
    and adds a few controls specific to the :class:`.HistogramPanel`.
    """



    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``HistogramToolBar`` is only intended to be added to
        :class:`.HistogramPanel` views.
        """

        return [histogrampanel.HistogramPanel]


    def __init__(self, parent, overlayList, displayCtx, histPanel):
        """Create a ``HistogramToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg histPanel:   The :class:`.HistogramPanel` instance.
        """

        plottoolbar.PlotToolBar.__init__(
            self, parent, overlayList, displayCtx, histPanel)

        togControl = actions.ToggleActionButton(
            'HistogramControlPanel',
            actionKwargs={'floatPane' : True},
            icon=[icons.findImageFile('spannerHighlight24'),
                  icons.findImageFile('spanner24')],
            tooltip=tooltips.actions[histPanel, 'HistogramControlPanel'])

        togList = actions.ToggleActionButton(
            'PlotListPanel',
            actionKwargs={'floatPane' : True},
            icon=[icons.findImageFile('listHighlight24'),
                  icons.findImageFile('list24')],
            tooltip=tooltips.actions[histPanel, 'PlotListPanel'])

        togOverlay = actions.ToggleActionButton(
            'toggleHistogramOverlay',
            icon=[icons.findImageFile('histogramOverlayHighlight24'),
                  icons.findImageFile('histogramOverlay24')])

        mode    = props.Widget(
            'histType',
            labels=strings.choices[     histPanel, 'histType'],
            tooltip=tooltips.properties[histPanel, 'histType'])


        togControl = props.buildGUI(self, histPanel, togControl)
        togList    = props.buildGUI(self, histPanel, togList)
        togOverlay = props.buildGUI(self, histPanel, togOverlay)
        mode       = props.buildGUI(self, histPanel, mode)

        lblMode = self.MakeLabelledTool(
            mode, strings.properties[histPanel, 'histType'])

        self.InsertTools([togControl, togList], 0)
        self.AddTool(togOverlay)
        self.AddTool(lblMode)


        nav = [togControl, togList]    + \
              self.getCommonNavOrder() + \
              [togOverlay, mode]

        self.setNavOrder(nav)



#!/usr/bin/env python
#
# loadvertexdata.py - The LoadVertexDataAction.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`LoadVertexDataAction`, which allows the
user to load additional vertex data or vertex sets for a :class:`.Mesh`
overlay. Two standalone functions, :func:`loadVertexData` and
:func:`loadVertices` are also provided.
"""


import os.path as op

import fsl.data.mesh                as fslmesh
import fsleyes_widgets.utils.status as status
import fsleyes.strings              as strings
from . import                          base


class LoadVertexDataAction(base.NeedOverlayAction):
    """The ``LoadVertexDataAction`` prompts the user to load a file containing
    vertex data or a vertex set for a :class:`.Mesh` overlay.  See the
    :attr:`.MeshOpts.vertexData` and :attr:`.MeshOpts.vertexSet` properties.
    """


    def __init__(self, overlayList, displayCtx, vertices=False):
        """Create a ``LoadVertexDataAction``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg vertices:    If ``True``, the user is prompted to load a file
                          containing vertices for the mesh. Otherwise, the user
                          is prompted to load a file containing vertex data.
        """
        base.NeedOverlayAction.__init__(
            self, overlayList, displayCtx, self.__onRun, fslmesh.Mesh)
        self.__vertices = vertices


    def __onRun(self):
        """Called when this action is executed.  Calls either
        :meth:`__loadVertexData`, or :meth:`__loadVertices`.
        """
        if self.__vertices: self.__loadVertices()
        else:               self.__loadVertexData()


    def __loadVertices(self):
        """Prompts the user to load a vertex file for the currently
        selected :class:`.Mesh` overlay, then sets the
        :attr:`.MeshOpts.vertexSet` property accordingly. If the file was
        successfully loaded, also adds the loaded file as an option on the
        :attr:`.MeshOpts.vertexSet` property.
        """
        self.__loadit('loadVertices', loadVertices)


    def __loadVertexData(self):
        """Prompts the user to load a vertex data file for the currently
        selected :class:`.Mesh` overlay, then sets the
        :attr:`.MeshOpts.vertexData` property accordingly. If the file was
        successfully loaded, also adds the loaded file as an option on the
        :attr:`.MeshOpts.vertexData` property.
        """
        self.__loadit('loadVertexData', loadVertexData)


    def __loadit(self, key, func):
        """Shared by the :meth:`__loadVertices` and :meth:`__loadVertexData`
        methods.
        """

        import wx

        app     = wx.GetApp()
        overlay = self.displayCtx.getSelectedOverlay()
        fromDir = op.dirname(overlay.dataSource)

        msg = strings.messages[self, key].format(overlay.name)
        dlg = wx.FileDialog(app.GetTopWindow(),
                            message=msg,
                            defaultDir=fromDir,
                            wildcard='*.*',
                            style=wx.FD_OPEN)

        if dlg.ShowModal() != wx.ID_OK:
            return

        path     = dlg.GetPath()
        errtitle = strings.titles[  self, 'error']
        errmsg   = strings.messages[self, 'error'].format(overlay.name)

        with status.reportIfError(errtitle, errmsg, raiseError=False):
            func(overlay, self.displayCtx, path)


def loadVertexData(overlay, displayCtx, filename, select=True):
    """Attempt to load the specified vertex data for the given overlay.

    :arg overlay:    The overlay (assumed to be a :class:`.Mesh` instance)

    :arg displayCtx: The :class:`.DisplayContext`

    :arg filename:   Path to the vertex data file that is to be loaded, or key
                     for vertex data that is already loaded (see the
                     :class:`.Mesh` class).

    :arg select:     If ``True`` (the default), the
                     :attr:`.MeshOpts.vertexData` is set to the
                     newly loaded file.

    :returns:        The path that was actually used - it will have been
                     converted to an absolute path if necessary.
    """

    opts     = displayCtx.getOpts(overlay)
    filename = op.abspath(filename)

    # The sole reason that this function exists is because
    # MeshOpts.vertexData is a props.Choice property, which
    # can only take one of a fixed set of values. So when
    # we want to load some data from a file that is not in
    # the possible values, we need to add the file as an
    # option before selecting it. A bit silly.
    if filename not in overlay.vertexDataSets():
        # Force the overlay to load
        # the vertex data. This will
        # throw an error if the file
        # is unrecognised.
        overlay.loadVertexData(filename)

        # Add the file as an
        # option, then select it.
        opts.addVertexDataOptions([filename])

    if select:
        opts.vertexData = filename

    return filename


def loadVertices(overlay, displayCtx, filename, select=True):
    """Attempt to load the specified vertexz file for the given overlay.

    :arg overlay:    The overlay (assumed to be a :class:`.Mesh` instance)

    :arg displayCtx: The :class:`.DisplayContext`

    :arg filename:   Path to the vertex file that is to be loaded, or key
                     for vertex data that is already loaded (see the
                     :class:`.Mesh` class).

    :arg select:     If ``True`` (the default), the
                     :attr:`.MeshOpts.vertexSet` is set to the
                     newly loaded file.

    :returns:        The path that was actually used - it will have been
                     converted to an absolute path if necessary.
    """

    opts     = displayCtx.getOpts(overlay)
    filename = op.abspath(filename)

    # We follow the same process
    # as in loadVertexData above
    if filename not in overlay.vertexSets():
        overlay.loadVertices(filename, select=False)
        opts.addVertexSetOptions([filename])

    if select:
        overlay.vertices = filename

    return filename



#!/usr/bin/env python
#
# __init__.py - Manage interactions between the FSLeyes FileTreePanel and
#               the filetree library.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``fsleyes.filetree`` package manages interactions between the FSLeyes
:class:`.FileTreePanel` and the `file-tree
<https://git.fmrib.ox.ac.uk/ndcn0236/file-tree/>`_ library.
"""


import os.path as op
import            glob

try:
    import file_tree
except ImportError:
    file_tree = None

from .filetreepanel import FileTreePanel
from .manager       import FileTreeManager
from .query         import FileTreeQuery


def list_all_trees():
    """Returns a list containing the names of all known ``.tree`` files. """

    file_tree.parse_tree.scan_plugins()

    treefiles = []

    for directory in file_tree.tree_directories:
        treefiles.extend(glob.glob(op.join(directory, '*.tree')))

    for subtree in file_tree.parse_tree.available_subtrees.keys():
        treefiles.append(subtree)

    return treefiles


def read(treefile, directory):
    """Load a ``.tree`` file, and return a ``FileTree``. """
    return file_tree.FileTree.read(treefile, top_level=directory)



#!/usr/bin/env python
#
# __init__.py - Classes used by PlotPanel.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``plotting`` package contains the :class:`.DataSeries` class, and
all of its sub-classes. These classes are used by :class:`.PlotPanel` views
for plotting data.
"""


from . import dataseries
from . import timeseries
from . import histogramseries
from . import powerspectrumseries

DataSeries                   = dataseries         .DataSeries
VoxelDataSeries              = dataseries         .VoxelDataSeries
VoxelTimeSeries              = timeseries         .VoxelTimeSeries
ComplexTimeSeries            = timeseries         .ComplexTimeSeries
ImaginaryTimeSeries          = timeseries         .ImaginaryTimeSeries
MagnitudeTimeSeries          = timeseries         .MagnitudeTimeSeries
PhaseTimeSeries              = timeseries         .PhaseTimeSeries
FEATTimeSeries               = timeseries         .FEATTimeSeries
FEATPartialFitTimeSeries     = timeseries         .FEATPartialFitTimeSeries
FEATEVTimeSeries             = timeseries         .FEATEVTimeSeries
FEATResidualTimeSeries       = timeseries         .FEATResidualTimeSeries
FEATModelFitTimeSeries       = timeseries         .FEATModelFitTimeSeries
MelodicTimeSeries            = timeseries         .MelodicTimeSeries
MeshTimeSeries               = timeseries         .MeshTimeSeries
HistogramSeries              = histogramseries    .HistogramSeries
ImageHistogramSeries         = histogramseries    .ImageHistogramSeries
MeshHistogramSeries          = histogramseries    .MeshHistogramSeries
PowerSpectrumSeries          = powerspectrumseries.PowerSpectrumSeries
VoxelPowerSpectrumSeries     = powerspectrumseries.VoxelPowerSpectrumSeries
ComplexPowerSpectrumSeries   = powerspectrumseries.ComplexPowerSpectrumSeries
ImaginaryPowerSpectrumSeries = powerspectrumseries.ImaginaryPowerSpectrumSeries
MagnitudePowerSpectrumSeries = powerspectrumseries.MagnitudePowerSpectrumSeries
PhasePowerSpectrumSeries     = powerspectrumseries.PhasePowerSpectrumSeries
MelodicPowerSpectrumSeries   = powerspectrumseries.MelodicPowerSpectrumSeries
MeshPowerSpectrumSeries      = powerspectrumseries.MeshPowerSpectrumSeries



#!/usr/bin/env python
#
# overlaydisplaywidgets.py - Contents of the OverlayDisplayPanel.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module is used by the :class:`.OverlayDisplayPanel`. It contains
definitions of all the settings that are displayed on the
``OverlayDisplayPanel`` for each overlay type.

It also contains functions which create customised widgets, for scenarios
where a widget does not directly map to a :class:`.Display` or
:class:`.DisplayOpts` property.


``_initPropertyList_[DisplayOptsType]``
``_init3DPropertyList_[DisplayOptsType]``
``_initWidgetSpec_[DisplayOptsType]``
``_init3DWidgetSpec_[DisplayOptsType]``
"""


import os.path as op
import            sys
import            copy
import            functools

import            wx

import fsl.utils.idle                 as idle
import fsleyes_props                  as props
import fsleyes_widgets                as fwidgets
import fsleyes_widgets.imagepanel     as imagepanel
import fsleyes_widgets.utils.typedict as td
import fsleyes.strings                as strings
import fsleyes.colourmaps             as fslcm
import fsleyes.controls.colourbar     as cbar
import fsleyes.actions.loadcolourmap  as loadcmap
import fsleyes.actions.loadvertexdata as loadvdata


_PROPERTIES      = td.TypeDict()
_3D_PROPERTIES   = td.TypeDict()
_WIDGET_SPECS    = td.TypeDict()
_3D_WIDGET_SPECS = td.TypeDict()




def _merge_dicts(d1, d2):
    d3 = d1.copy()
    d3.update(d2)
    return d3


def getPropertyList(target, threedee=False):

    plist = _getThing(target, '_initPropertyList_', _PROPERTIES, threedee)

    if plist is None:
        return []

    return functools.reduce(lambda a, b: a + b, plist)


def get3DPropertyList(target):

    plist = _getThing(target, '_init3DPropertyList_', _3D_PROPERTIES)

    if plist is None:
        return []

    return functools.reduce(lambda a, b: a + b, plist)


def getWidgetSpecs(target, threedee=False):

    sdicts = _getThing(target, '_initWidgetSpec_', _WIDGET_SPECS, threedee)

    if sdicts is None:
        return {}

    return functools.reduce(_merge_dicts, sdicts)


def get3DWidgetSpecs(target):

    sdicts = _getThing(target, '_init3DWidgetSpec_', _3D_WIDGET_SPECS)

    if sdicts is None:
        return {}

    return functools.reduce(_merge_dicts, sdicts)


def _getThing(target, prefix, thingDict, *args, **kwargs):

    def _makeKey(t):
        return (t, str(args), str(kwargs))

    tkey = _makeKey(target)

    if thingDict.get(tkey, None, exact=True) is None:

        keys, funcs = _getInitFuncs(prefix, target)

        for key, func in zip(keys, funcs):
            key = _makeKey(key)
            thingDict[key] = func(*args, **kwargs)

    return thingDict.get(tkey, None, allhits=True)


def _getInitFuncs(prefix, target):

    if isinstance(target, type): ttype = target
    else:                        ttype = type(target)

    key   = ttype.__name__
    bases = ttype.__bases__

    thismod  = sys.modules[__name__]
    initFunc = '{}{}'.format(prefix, key)
    initFunc = getattr(thismod, initFunc, None)

    if initFunc is None:
        return [], []

    keys      = [key]
    initFuncs = [initFunc]

    for base in bases:
        bkeys, bfuncs = _getInitFuncs(prefix, base)

        keys     .extend(bkeys)
        initFuncs.extend(bfuncs)

    return keys, initFuncs


def _initPropertyList_Display(threedee):
    return ['name',
            'overlayType',
            'enabled',
            'alpha',
            'brightness',
            'contrast']


def _initPropertyList_VolumeOpts(threedee):
    plist = ['custom_volume',
             'channel',
             'interpolation',
             'custom_cmap',
             'cmapResolution',
             'gamma',
             'interpolateCmaps',
             'invert',
             'invertClipping',
             'linkLowRanges',
             'linkHighRanges',
             'modulateAlpha',
             'clipImage',
             'modulateImage',
             'displayRange',
             'clippingRange',
             'modulateRange',
             'custom_overrideDataRange']

    if threedee:
        plist.remove('modulateAlpha')
        plist.remove('clipImage')
        plist.remove('modulateImage')
        plist.remove('modulateRange')

    return plist


def _initPropertyList_ComplexOpts(threedee):
    return ['component']


def _init3DPropertyList_VolumeOpts():
    return ['blendFactor',
            'blendByIntensity',
            'numSteps',
            'resolution',
            'custom_clipPlanes']


def _initPropertyList_MaskOpts(threedee):
    return ['custom_volume',
            'colour',
            'invert',
            'threshold',
            'interpolation',
            'outline',
            'outlineWidth']


def _initPropertyList_VectorOpts(threedee):
    return ['colourImage',
            'modulateImage',
            'clipImage',
            'cmap',
            'clippingRange',
            'modulateRange',
            'modulateMode',
            'xColour',
            'yColour',
            'zColour',
            'suppressX',
            'suppressY',
            'suppressZ',
            'suppressMode']


def _initPropertyList_RGBVectorOpts(threedee):
    return ['interpolation',
            'unitLength']


def _initPropertyList_LineVectorOpts(threedee):
    return ['directed',
            'unitLength',
            'orientFlip',
            'lineWidth',
            'lengthScale']


def _initPropertyList_TensorOpts(threedee):
    return ['lighting',
            'orientFlip',
            'tensorResolution',
            'tensorScale']


def _initPropertyList_MeshOpts(threedee):
    plist = ['refImage',
             'coordSpace',
             'outline',
             'outlineWidth',
             'colour',
             'custom_vertexSet',
             'custom_vertexData',
             'vertexDataIndex',
             'custom_lut',
             'custom_cmap',
             'cmapResolution',
             'gamma',
             'interpolateCmaps',
             'invert',
             'invertClipping',
             'discardClipped',
             'linkLowRanges',
             'linkHighRanges',
             'modulateAlpha',
             'modulateData',
             'displayRange',
             'clippingRange',
             'modulateRange']

    # Remove outline
    # options for 3D
    if threedee:
        plist.remove('outlineWidth')
        plist.remove('outline')

    return plist


def _init3DPropertyList_MeshOpts():
    return ['flatShading', 'wireframe']


def _initPropertyList_GiftiOpts(threedee):
    return []


def _init3DPropertyList_GiftiOpts():
    return []


def _initPropertyList_FreesurferOpts(threedee):
    return []


def _init3DPropertyList_FreesurferOpts():
    return []


def _initPropertyList_LabelOpts(threedee):
    return ['lut',
            'outline',
            'outlineWidth',
            'custom_volume']


def _initPropertyList_SHOpts(threedee):
    return ['shResolution',
            'shOrder',
            'orientFlip',
            'lighting',
            'normalise',
            'size',
            'radiusThreshold',
            'colourMode']


def _initPropertyList_MIPOpts(threedee):
    return ['custom_volume',
            'interpolation',
            'cmap',
            'cmapResolution',
            'gamma',
            'interpolateCmaps',
            'invert',
            'invertClipping',
            'linkLowRanges',
            'linkHighRanges',
            'displayRange',
            'clippingRange',
            'window',
            'minimum',
            'absolute']


def _initPropertyList_VolumeRGBOpts(threedee):
    return ['custom_volume',
            'interpolation',
            'rColour',
            'gColour',
            'bColour',
            'suppressR',
            'suppressG',
            'suppressB',
            'suppressA',
            'suppressMode']


def _initWidgetSpec_Display(threedee):
    return {
        'name'        : props.Widget('name'),
        'overlayType' : props.Widget(
            'overlayType',
            labels=strings.choices['Display.overlayType']),
        'enabled'     : props.Widget('enabled'),
        'alpha'       : props.Widget('alpha',      showLimits=False),
        'brightness'  : props.Widget('brightness', showLimits=False),
        'contrast'    : props.Widget('contrast',   showLimits=False),
    }


def _initWidgetSpec_ColourMapOpts(threedee):
    return {
        'custom_cmap'     : _ColourMapOpts_ColourMapWidget,
        'cmap'            : props.Widget(
            'cmap',
            labels=fslcm.getColourMapLabel),
        'useNegativeCmap' : props.Widget('useNegativeCmap'),
        'negativeCmap'    : props.Widget(
            'negativeCmap',
            labels=fslcm.getColourMapLabel,
            dependencies=['useNegativeCmap'],
            enabledWhen=lambda i, unc : unc),
        'cmapResolution'  : props.Widget(
            'cmapResolution',
            slider=True,
            spin=True,
            showLimits=False),
        'interpolateCmaps' : props.Widget('interpolateCmaps'),
        'invert'           : props.Widget('invert'),
        'invertClipping'   : props.Widget('invertClipping'),
        'linkLowRanges'    : props.Widget('linkLowRanges'),
        'linkHighRanges'   : props.Widget('linkHighRanges'),
        'modulateAlpha'    : props.Widget('modulateAlpha'),
        'gamma'            : props.Widget(
            'gamma',
            showLimits=False,
            slider=True,
            spin=True),
        'displayRange'     : props.Widget(
            'displayRange',
            showLimits=False,
            slider=True,
            labels=[strings.choices['ColourMapOpts.displayRange.min'],
                    strings.choices['ColourMapOpts.displayRange.max']]),
        'clippingRange'  : props.Widget(
            'clippingRange',
            showLimits=False,
            slider=True,
            labels=[strings.choices['ColourMapOpts.displayRange.min'],
                    strings.choices['ColourMapOpts.displayRange.max']]),
        'modulateRange'  : props.Widget(
            'modulateRange',
            showLimits=False,
            slider=True,
            dependencies=['modulateAlpha'],
            enabledWhen=lambda i, ma : ma,
            labels=[strings.choices['ColourMapOpts.displayRange.min'],
                    strings.choices['ColourMapOpts.displayRange.max']]),
    }


def _initWidgetSpec_VolumeOpts(threedee):

    def imageName(img):
        if img is None: return 'None'
        else:           return img.name

    return {
        'custom_volume'            : _NiftiOpts_VolumeWidget,
        'custom_overrideDataRange' : _VolumeOpts_OverrideDataRangeWidget,

        'channel'        : props.Widget('channel'),
        'volume'         : props.Widget(
            'volume',
            showLimits=False,
            enabledWhen=lambda o: o.overlay.ndim > 3,
            spinWidth=6),
        'volumeDim'      : props.Widget(
            'volumeDim',
            showLimits=False,
            slider=False,
            enabledWhen=lambda o: o.overlay.ndim > 4,
            spinWidth=2),
        'interpolation'  : props.Widget(
            'interpolation',
            labels=strings.choices['VolumeOpts.interpolation']),
        'clipImage'      : props.Widget(
            'clipImage',
            labels=imageName),
        'modulateImage'      : props.Widget(
            'modulateImage',
            labels=imageName),
        'enableOverrideDataRange'  : props.Widget(
            'enableOverrideDataRange'),
        'overrideDataRange' : props.Widget(
            'overrideDataRange',
            showLimits=False,
            spin=True,
            slider=False,
            dependencies=['enableOverrideDataRange'],
            enabledWhen=lambda vo, en: en),
    }


def _init3DWidgetSpec_VolumeOpts():

    return {
        'numSteps'          : props.Widget('numSteps',
                                           showLimits=False),
        'blendFactor'       : props.Widget('blendFactor',
                                           showLimits=False),
        'blendByIntensity'  : props.Widget('blendByIntensity',
                                           showLimits=False),
        'resolution'        : props.Widget('resolution',
                                           showLimits=False),
        'numClipPlanes'     : props.Widget('numClipPlanes',
                                           slider=False,
                                           showLimits=False),
        'clipMode'          : props.Widget(
            'clipMode',
            labels=strings.choices['Volume3DOpts.clipMode']),
        'showClipPlanes'    : props.Widget('showClipPlanes'),
        'clipPosition'      : props.Widget('clipPosition',
                                           showLimits=False),
        'clipAzimuth'       : props.Widget('clipAzimuth',
                                           showLimits=False),
        'clipInclination'   : props.Widget('clipInclination',
                                           showLimits=False),
        'custom_clipPlanes' : _VolumeOpts_3DClipPlanes,
    }


def _initWidgetSpec_ComplexOpts(threedee):
    return {
        'component' : props.Widget(
            'component',
            labels=strings.choices['ComplexOpts.component'])
    }


def _initWidgetSpec_MaskOpts(threedee):
    return {
        'custom_volume'  : _NiftiOpts_VolumeWidget,
        'volume'         : props.Widget(
            'volume',
            showLimits=False,
            enabledWhen=lambda o: o.overlay.ndim > 3,
            spinWidth=6),
        'volumeDim'      : props.Widget(
            'volumeDim',
            showLimits=False,
            slider=False,
            enabledWhen=lambda o: o.overlay.ndim > 4,
            spinWidth=2),
        'interpolation'  : props.Widget(
            'interpolation',
            labels=strings.choices['VolumeOpts.interpolation']),
        'colour'       : props.Widget('colour'),
        'invert'       : props.Widget('invert'),
        'threshold'    : props.Widget('threshold', showLimits=False),
        'outline'      : props.Widget('outline'),
        'outlineWidth' : props.Widget('outlineWidth', showLimits=False),
    }


def _initWidgetSpec_LabelOpts(threedee):
    return {
        'lut'          : props.Widget('lut', labels=lambda l: l.name),
        'outline'      : props.Widget('outline'),
        'outlineWidth' : props.Widget('outlineWidth', showLimits=False),
        'custom_volume'  : _NiftiOpts_VolumeWidget,
        'volume'         : props.Widget(
            'volume',
            showLimits=False,
            enabledWhen=lambda o: o.overlay.ndim > 3,
            spinWidth=6),
        'volumeDim'      : props.Widget(
            'volumeDim',
            showLimits=False,
            slider=False,
            enabledWhen=lambda o: o.overlay.ndim > 4,
            spinWidth=2),
    }


def _initWidgetSpec_VectorOpts(threedee):
    def imageName(img):
        if img is None: return 'None'
        else:           return img.name

    return {
        'colourImage'   : props.Widget(
            'colourImage',
            labels=imageName),
        'modulateImage' : props.Widget(
            'modulateImage',
            labels=imageName,
            dependencies=['colourImage'],
            enabledWhen=lambda o, ci: ci is None),
        'clipImage'     : props.Widget('clipImage', labels=imageName),
        'cmap'          : props.Widget(
            'cmap',
            labels=fslcm.getColourMapLabel,
            dependencies=['colourImage'],
            enabledWhen=lambda o, ci: ci is not None),
        'clippingRange' : props.Widget(
            'clippingRange',
            showLimits=False,
            slider=True,
            labels=[strings.choices['VectorOpts.clippingRange.min'],
                    strings.choices['VectorOpts.clippingRange.max']],
            dependencies=['clipImage'],
            enabledWhen=lambda o, ci: ci is not None),
        'modulateRange' : props.Widget(
            'modulateRange',
            showLimits=False,
            slider=True,
            labels=[strings.choices['VectorOpts.modulateRange.min'],
                    strings.choices['VectorOpts.modulateRange.max']],
            dependencies=['colourImage', 'modulateImage'],
            enabledWhen=lambda o, ci, mi: ci is None and mi is not None),
        'modulateMode' : props.Widget(
            'modulateMode',
            labels=strings.choices['VectorOpts.modulateMode'],
            dependencies=['colourImage', 'modulateImage'],
            enabledWhen=lambda o, ci, mi: ci is None and mi is not None),
        'xColour'       : props.Widget(
            'xColour',
            dependencies=['colourImage'],
            enabledWhen=lambda o, ci: ci is None),
        'yColour'       : props.Widget(
            'yColour',
            dependencies=['colourImage'],
            enabledWhen=lambda o, ci: ci is None),
        'zColour'       : props.Widget(
            'zColour',
            dependencies=['colourImage'],
            enabledWhen=lambda o, ci: ci is None),
        'suppressX'     : props.Widget(
            'suppressX',
            dependencies=['colourImage'],
            enabledWhen=lambda o, ci: ci is None),
        'suppressY'     : props.Widget(
            'suppressY',
            dependencies=['colourImage'],
            enabledWhen=lambda o, ci: ci is None),
        'suppressZ'     : props.Widget(
            'suppressZ',
            dependencies=['colourImage'],
            enabledWhen=lambda o, ci: ci is None),
        'suppressMode'  : props.Widget(
            'suppressMode',
            dependencies=['colourImage'],
            labels=strings.choices['VectorOpts.suppressMode'],
            enabledWhen=lambda o, ci: ci is None),
    }


def _initWidgetSpec_RGBVectorOpts(threedee):
    return {
        'interpolation' : props.Widget(
            'interpolation',
            labels=strings.choices['VolumeOpts.interpolation']),
        'unitLength' : props.Widget('unitLength')
    }


def _initWidgetSpec_LineVectorOpts(threedee):
    return {
        'directed'    : props.Widget('directed'),
        'unitLength'  : props.Widget('unitLength'),
        'orientFlip'  : props.Widget('orientFlip'),
        'lineWidth'   : props.Widget('lineWidth',   showLimits=False),
        'lengthScale' : props.Widget('lengthScale', showLimits=False),
    }


def _initWidgetSpec_TensorOpts(threedee):
    return {
        'lighting'         : props.Widget('lighting'),
        'orientFlip'       : props.Widget('orientFlip'),
        'tensorResolution' : props.Widget(
            'tensorResolution',
            showLimits=False,
            spin=False,
            labels=[strings.choices['TensorOpts.tensorResolution.min'],
                    strings.choices['TensorOpts.tensorResolution.max']]),
        'tensorScale'      : props.Widget(
            'tensorScale',
            showLimits=False,
            spin=False),
    }

def _initWidgetSpec_SHOpts(threedee):
    return {
        'shResolution'    : props.Widget(
            'shResolution',
            spin=False,
            showLimits=False),
        'shOrder'    : props.Widget('shOrder'),
        'orientFlip' : props.Widget('orientFlip'),
        'lighting'   : props.Widget('lighting'),
        'normalise'  : props.Widget('normalise'),
        'size'       : props.Widget(
            'size',
            spin=False,
            showLimits=False),
        'radiusThreshold' : props.Widget(
            'radiusThreshold',
            spin=False,
            showLimits=False),
        'colourMode'      : props.Widget(
            'colourMode',
            labels=strings.choices['SHOpts.colourMode'],
            dependencies=['colourImage'],
            enabledWhen=lambda o, ci: ci is None),
        'cmap' : props.Widget(
            'cmap',
            labels=fslcm.getColourMapLabel,
            dependencies=['colourImage', 'colourMode'],
            enabledWhen=lambda o, ci, cm: ci is not None or cm == 'radius'),
        'xColour'         : props.Widget(
            'xColour',
            dependencies=['colourImage', 'colourMode'],
            enabledWhen=lambda o, ci, cm: ci is None and cm == 'direction'),
        'yColour'         : props.Widget(
            'yColour',
            dependencies=['colourImage', 'colourMode'],
            enabledWhen=lambda o, ci, cm: ci is None and cm == 'direction'),
        'zColour'         : props.Widget(
            'zColour',
            dependencies=['colourImage', 'colourMode'],
            enabledWhen=lambda o, ci, cm: ci is None and cm == 'direction'),
        'suppressX'         : props.Widget(
            'suppressX',
            dependencies=['colourImage', 'colourMode'],
            enabledWhen=lambda o, ci, cm: ci is None and cm == 'direction'),
        'suppressY'         : props.Widget(
            'suppressY',
            dependencies=['colourImage', 'colourMode'],
            enabledWhen=lambda o, ci, cm: ci is None and cm == 'direction'),
        'suppressZ'         : props.Widget(
            'suppressZ',
            dependencies=['colourImage', 'colourMode'],
            enabledWhen=lambda o, ci, cm: ci is None and cm == 'direction'),
        'suppressMode'         : props.Widget(
            'suppressMode',
            dependencies=['colourImage', 'colourMode'],
            enabledWhen=lambda o, ci, cm: ci is None and cm == 'direction'),
    }



def _initWidgetSpec_MeshOpts(threedee):

    def imageName(img):
        if img is None: return 'None'
        else:           return img.name

    def pathName(vdata):
        if vdata is None: return 'None'
        else:             return op.basename(vdata)

    def colourEnabledWhen(opts, vdata, useLut):
        return (vdata is not None) and (not useLut)

    colourKwargs = {
        'dependencies' : ['vertexData', 'useLut'],
        'enabledWhen'  : colourEnabledWhen
    }

    return {
        'outline'      : props.Widget(
            'outline',
            dependencies=['vertexData'],
            enabledWhen=lambda o, v: v is None),
        'outlineWidth' : props.Widget(
            'outlineWidth',
            showLimits=False,
            dependencies=['outline', 'vertexData'],
            enabledWhen=lambda op, o, v: o or v is not None),
        'refImage'     : props.Widget('refImage', labels=imageName),
        'coordSpace'   : props.Widget(
            'coordSpace',
            enabledWhen=lambda o, ri: ri != 'none',
            labels=strings.choices['MeshOpts.coordSpace'],
            dependencies=['refImage']),
        'colour'       : props.Widget('colour'),
        'custom_vertexData' : _MeshOpts_vertexDataWidget,
        'custom_vertexSet'  : _MeshOpts_vertexSetWidget,
        'vertexSet'    : props.Widget(
            'vertexSet',
            labels=pathName),
        'vertexData'   : props.Widget(
            'vertexData',
            labels=pathName),
        'modulateData'   : props.Widget(
            'modulateData',
            labels=pathName),
        'vertexDataIndex' : props.Widget(
            'vertexDataIndex',
            showLimits=False,
            dependencies=['vertexData'],
            enabledWhen=lambda o, vd: vd is not None),
        'useLut' : props.Widget(
            'useLut',
            dependencies=['vertexData'],
            enabledWhen=lambda o, vd: vd is not None),
        'custom_lut' : _MeshOpts_LutWidget,
        'lut'    : props.Widget(
            'lut',
            labels=lambda l: l.name,
            dependencies=['vertexData'],
            enabledWhen=lambda o, vd: vd is not None),

        # We override the ColourMapOpts definitions
        # for custom enabledWhen behaviour.
        'cmap'           : props.Widget(
            'cmap',
            labels=fslcm.getColourMapLabel,
            **colourKwargs),

        'useNegativeCmap' : props.Widget(
            'useNegativeCmap',
            **colourKwargs),
        'negativeCmap'    : props.Widget(
            'negativeCmap',
            labels=fslcm.getColourMapLabel,
            **colourKwargs),
        'cmapResolution'  : props.Widget(
            'cmapResolution',
            slider=True,
            spin=True,
            showLimits=False,
            **colourKwargs),
        'interpolateCmaps' : props.Widget(
            'interpolateCmaps',
            **colourKwargs),
        'invert'           : props.Widget(
            'invert',
            **colourKwargs),
        'invertClipping'   : props.Widget(
            'invertClipping',
            **colourKwargs),
        'linkLowRanges'    : props.Widget(
            'linkLowRanges',
            **colourKwargs),
        'linkHighRanges' : props.Widget(
            'linkHighRanges',
            **colourKwargs),
        'displayRange'   : props.Widget(
            'displayRange',
            showLimits=False,
            slider=True,
            labels=[strings.choices['ColourMapOpts.displayRange.min'],
                    strings.choices['ColourMapOpts.displayRange.max']],
            **colourKwargs),
        'clippingRange'  : props.Widget(
            'clippingRange',
            showLimits=False,
            slider=True,
            labels=[strings.choices['ColourMapOpts.displayRange.min'],
                    strings.choices['ColourMapOpts.displayRange.max']],
            dependencies=['vertexData'],
            enabledWhen=lambda o, vd: vd is not None),
        'discardClipped' : props.Widget(
            'discardClipped',
            **colourKwargs),
    }


def _init3DWidgetSpec_MeshOpts():
    return {
        'wireframe' : props.Widget('wireframe'),
        'flatShading' : props.Widget(
            'flatShading',
            dependencies=['vertexData'],
            enabledWhen=lambda o, vd: vd is not None),
    }


def _initWidgetSpec_GiftiOpts(threedee):
    return {}
def _init3DWidgetSpec_GiftiOpts():
    return {}


def _initWidgetSpec_FreesurferOpts(threedee):
    return {}
def _init3DWidgetSpec_FreesurferOpts():
    return {}


def _initWidgetSpec_MIPOpts(threedee):

    return {
        'custom_volume' : _NiftiOpts_VolumeWidget,
        'volume'         : props.Widget(
            'volume',
            showLimits=False,
            enabledWhen=lambda o: o.overlay.ndim > 3,
            spinWidth=6),
        'volumeDim'      : props.Widget(
            'volumeDim',
            showLimits=False,
            slider=False,
            enabledWhen=lambda o: o.overlay.ndim > 4,
            spinWidth=2),
        'interpolation'  : props.Widget(
            'interpolation',
            labels=strings.choices['VolumeOpts.interpolation']),
        'window' : props.Widget(
            'window',
            showLimits=False,
            slider=True),
        'minimum'  : props.Widget('minimum'),
        'absolute' : props.Widget('absolute'),
    }


def _initWidgetSpec_VolumeRGBOpts(threedee):

    return {
        'custom_volume'  : _NiftiOpts_VolumeWidget,
        'volume'         : props.Widget(
            'volume',
            showLimits=False,
            enabledWhen=lambda o: o.overlay.ndim > 3,
            spinWidth=6),
        'volumeDim'      : props.Widget(
            'volumeDim',
            showLimits=False,
            slider=False,
            enabledWhen=lambda o: o.overlay.ndim > 4,
            spinWidth=2),
        'interpolation'  : props.Widget(
            'interpolation',
            labels=strings.choices['VolumeOpts.interpolation']),
        'rColour'      : props.Widget('rColour'),
        'gColour'      : props.Widget('gColour'),
        'bColour'      : props.Widget('bColour'),
        'suppressR'    : props.Widget('suppressR'),
        'suppressG'    : props.Widget('suppressG'),
        'suppressB'    : props.Widget('suppressB'),
        'suppressA'    : props.Widget('suppressA'),
        'suppressMode' : props.Widget('suppressMode'),
    }


def _ColourMapOpts_ColourMapWidget(
        target,
        parent,
        panel,
        overlayList,
        displayCtx,
        threedee):
    """Builds a panel which contains widgets for controlling the
    :attr:`.ColourMapOpts.cmap`, :attr:`.ColourMapOpts.negativeCmap`, and
    :attr:`.ColourMapOpts.useNegativeCmap`.

    :returns: A ``wx.Sizer`` containing all of the widgets, and a list
              containing the extra widgets that were added.
    """

    # Button to load a new
    # colour map from file
    loadAction = loadcmap.LoadColourMapAction(overlayList, displayCtx)

    loadButton = wx.Button(parent)
    loadButton.SetLabel(strings.labels[panel, 'loadCmap'])

    loadAction.bindToWidget(panel, wx.EVT_BUTTON, loadButton)

    cmap       = getWidgetSpecs(target, threedee)['cmap']
    negCmap    = getWidgetSpecs(target, threedee)['negativeCmap']
    useNegCmap = getWidgetSpecs(target, threedee)['useNegativeCmap']

    cbpanel    = imagepanel.ImagePanel(parent)
    cbpanel.SetMinSize((-1, 30))
    colourbar  = cbar.ColourBar(overlayList, displayCtx)

    colourbar.bgColour  = (0, 0, 0, 0)
    colourbar.showLabel = False
    colourbar.showTicks = False

    def cbarUpdate(*a):
        w, h = cbpanel.GetSize().Get()

        if w < 20 or h < 20:
            return

        bmp = colourbar.colourBar(w, h)

        if bmp is None:
            return

        bmp = wx.Bitmap.FromBufferRGBA(w, h, bmp.transpose(1, 0, 2))

        cbpanel.SetImage(bmp.ConvertToImage())

    lname = 'ColourBarWidget_{}'.format(colourbar)

    def onDestroy(ev):
        colourbar.deregister(lname)
        colourbar.destroy()

    colourbar.register(lname,           cbarUpdate)
    cbpanel.Bind(wx.EVT_SIZE,           cbarUpdate)
    cbpanel.Bind(wx.EVT_WINDOW_DESTROY, onDestroy)

    cbarUpdate()

    cmap       = props.buildGUI(parent, target, cmap)
    negCmap    = props.buildGUI(parent, target, negCmap)
    useNegCmap = props.buildGUI(parent, target, useNegCmap)

    useNegCmap.SetLabel(strings.properties[target, 'useNegativeCmap'])

    sizer = wx.GridBagSizer()
    sizer.AddGrowableCol(0)

    sizer.Add(cbpanel,    (0, 0), (1, 2), flag=wx.EXPAND)
    sizer.Add(cmap,       (1, 0), (1, 1), flag=wx.EXPAND)
    sizer.Add(loadButton, (1, 1), (1, 1), flag=wx.EXPAND)
    sizer.Add(negCmap,    (2, 0), (1, 1), flag=wx.EXPAND)
    sizer.Add(useNegCmap, (2, 1), (1, 1), flag=wx.EXPAND)

    return sizer, [cmap, negCmap, useNegCmap]


def _NiftiOpts_VolumeWidget(
        target,
        parent,
        panel,
        overlayList,
        displayCtx,
        threedee):
    """Builds a panel which contains widgets for the
    :attr:`.NiftiOpts.volume` and :attr:`.NiftiOpts.volumeDim` properties.
    """

    volume    = getWidgetSpecs(target, threedee)['volume']
    volumeDim = getWidgetSpecs(target, threedee)['volumeDim']

    volume    = props.buildGUI(parent, target, volume)
    volumeDim = props.buildGUI(parent, target, volumeDim)

    dimLabel = wx.StaticText(parent,
                             label='Dim',
                             style=(wx.ALIGN_CENTRE_VERTICAL |
                                    wx.ALIGN_CENTRE_HORIZONTAL))
    dimLabel.SetMinSize(dimLabel.GetBestSize())
    sizer = wx.BoxSizer(wx.HORIZONTAL)

    sizer.Add(volume,    flag=wx.EXPAND, proportion=1)
    sizer.Add(dimLabel,  flag=wx.EXPAND)
    sizer.Add(volumeDim, flag=wx.EXPAND)

    return sizer, [volume, volumeDim]


def _VolumeOpts_OverrideDataRangeWidget(
        target,
        parent,
        panel,
        overlayList,
        displayCtx,
        threedee):
    """Builds a panel which contains widgets for enabling and adjusting
    the :attr:`.VolumeOpts.overrideDataRange`.

    :returns: a ``wx.Sizer`` containing all of the widgets.
    """

    # Override data range widget
    enable   = getWidgetSpecs(target, threedee)['enableOverrideDataRange']
    ovrRange = getWidgetSpecs(target, threedee)['overrideDataRange']

    enable   = props.buildGUI(parent, target, enable)
    ovrRange = props.buildGUI(parent, target, ovrRange)

    sizer = wx.BoxSizer(wx.HORIZONTAL)

    sizer.Add(enable,   flag=wx.EXPAND)
    sizer.Add(ovrRange, flag=wx.EXPAND, proportion=1)

    return sizer, [enable, ovrRange]


def _VolumeOpts_3DClipPlanes(
        target,
        parent,
        panel,
        overlayList,
        displayCtx,
        threedee):
    """Generates widget specifications for the ``VolumeOpts`` 3D settings.
    A different number of widgets are shown depending on the value of the
    :attr:`.VolumeOpts.numClipPlanes` setting.
    """

    # Whenever numClipPlanes changes, we
    # need to refresh the clip plane widgets.
    # Easiest way to do this is to tell the
    # OverlayDisplayPanel to re-create the 3D
    # settings section.
    #
    # TODO what is the lifespan of this listener?
    def numClipPlanesChanged(*a):
        if fwidgets.isalive(panel) and \
           fwidgets.isalive(parent):
            idle.idle(panel.updateWidgets, target, '3d')

    name = '{}_{}_VolumeOpts_3DClipPlanes'.format(
        target.name, id(panel))

    target.addListener('numClipPlanes',
                       name,
                       numClipPlanesChanged,
                       overwrite=True,
                       weak=False)

    numPlanes    = target.numClipPlanes
    numPlaneSpec = get3DWidgetSpecs(target)['numClipPlanes']
    clipMode     = get3DWidgetSpecs(target)['clipMode']
    showPlanes   = get3DWidgetSpecs(target)['showClipPlanes']
    position     = get3DWidgetSpecs(target)['clipPosition']
    azimuth      = get3DWidgetSpecs(target)['clipAzimuth']
    inclination  = get3DWidgetSpecs(target)['clipInclination']

    specs = [numPlaneSpec, showPlanes, clipMode]

    if numPlanes == 0:
        return specs, None

    positions    = [copy.deepcopy(position)    for i in range(numPlanes)]
    azimuths     = [copy.deepcopy(azimuth)     for i in range(numPlanes)]
    inclinations = [copy.deepcopy(inclination) for i in range(numPlanes)]

    for i in range(numPlanes):

        positions[i]   .index = i
        azimuths[i]    .index = i
        inclinations[i].index = i

        label = strings.labels[panel, 'clipPlane#'].format(i + 1)
        label = props.Label(label=label)

        specs.extend((label, positions[i], azimuths[i], inclinations[i]))

    return specs, None


def _MeshOpts_vertexDataWidget(
        target,
        parent,
        panel,
        overlayList,
        displayCtx,
        threedee):
    """Builds a panel which contains a widget for controlling the
    :attr:`.MeshOpts.vertexData` property, and also has a button
    which opens a file dialog, allowing the user to select other
    data.
    """

    loadAction = loadvdata.LoadVertexDataAction(overlayList, displayCtx)
    loadButton = wx.Button(parent)
    loadButton.SetLabel(strings.labels[panel, 'loadVertexData'])

    loadAction.bindToWidget(panel, wx.EVT_BUTTON, loadButton)

    sizer = wx.BoxSizer(wx.HORIZONTAL)

    vdata = getWidgetSpecs(target, threedee)['vertexData']
    vdata = props.buildGUI(parent, target, vdata)

    sizer.Add(vdata,      flag=wx.EXPAND, proportion=1)
    sizer.Add(loadButton, flag=wx.EXPAND)

    return sizer, [vdata]


def _MeshOpts_vertexSetWidget(
        target,
        parent,
        panel,
        overlayList,
        displayCtx,
        threedee):
    """Builds a panel which contains a widget for controlling the
    :attr:`.MeshOpts.vertexSet` property, and also has a button
    which opens a file dialog, allowing the user to select other
    data.
    """

    loadAction = loadvdata.LoadVertexDataAction(overlayList,
                                                displayCtx,
                                                vertices=True)
    loadButton = wx.Button(parent)
    loadButton.SetLabel(strings.labels[panel, 'loadVertices'])

    loadAction.bindToWidget(panel, wx.EVT_BUTTON, loadButton)

    sizer = wx.BoxSizer(wx.HORIZONTAL)

    vdata = getWidgetSpecs(target, threedee)['vertexSet']
    vdata = props.buildGUI(parent, target, vdata)

    sizer.Add(vdata,      flag=wx.EXPAND, proportion=1)
    sizer.Add(loadButton, flag=wx.EXPAND)

    return sizer, [vdata]


def _MeshOpts_LutWidget(
        target,
        parent,
        panel,
        overlayList,
        displayCtx,
        threedee):
    """Builds a panel which contains the provided :attr:`.MeshOpts.lut`
    widget, and also a widget for :attr:`.MeshOpts.useLut`.
    """

    # enable lut widget
    lut    = getWidgetSpecs(target, threedee)['lut']
    enable = getWidgetSpecs(target, threedee)['useLut']

    lut    = props.buildGUI(parent, target, lut)
    enable = props.buildGUI(parent, target, enable)

    sizer = wx.BoxSizer(wx.HORIZONTAL)

    sizer.Add(enable, flag=wx.EXPAND)
    sizer.Add(lut,    flag=wx.EXPAND, proportion=1)

    return sizer, [enable, lut]



#!/usr/bin/env python
#
# orthoeditactiontoolbar.py - Ortho edit mode action toolbar.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoEditActionToolbar`, a toolbar used
by the :class:`.OrthoPanel` in edit mode, which contains various buttons
allowing the user to run various edit-related actions.
"""

import wx

import fsleyes_props                           as props
import fsleyes.controls.controlpanel           as ctrlpanel
import fsleyes.controls.orthoeditsettingspanel as editpanel
import fsleyes.profiles.orthoeditprofile       as orthoeditprofile
import fsleyes.toolbar                         as fsltoolbar
import fsleyes.actions                         as actions
import fsleyes.icons                           as fslicons
import fsleyes.tooltips                        as fsltooltips


class OrthoEditActionToolBar(ctrlpanel.ControlToolBar):
    """The ``OrthoEditActionToolBar`` is a toolbar used by the
    :class:`.OrthoPanel`, which contains buttons allowing the user to:

     - Open the :class:`.OrthoEditSettingsPanel`
     - Create a new :class:`.Image`
     - Undo/redo the last change
     - Clear/fill/erase the current selection
     - Copy/paste data between images
    """


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``OrthoEditActionToolBar`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        from fsleyes.views.orthopanel import OrthoPanel
        return [OrthoPanel]


    @staticmethod
    def profileCls():
        """The ``OrthoEditActionToolBar`` is intended to be activated with the
        :class:`.OrthoEditProfile`.
        """
        return orthoeditprofile.OrthoEditProfile



    @staticmethod
    def ignoreControl():
        """The ``OrthoEditActionToolBar`` is not intended to be explicitly
        added by the user - it is added via :meth:`.OrthoPanel.toggleEditMode`.
        Overriding this method tells the :class:`.FSLeyesFrame` that it
        should not be added to the ortho panel settings menu.
        """
        return True


    def __init__(self, parent, overlayList, displayCtx, ortho):
        """Create an ``OrthoEditActionToolBar``.

        :arg parent:      The :mod:`wx` parent object
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          ortho,
                                          height=24,
                                          orient=wx.VERTICAL,
                                          kbFocus=True)

        self.toggleEditPanel = actions.ToggleControlPanelAction(
            overlayList, displayCtx, ortho, editpanel.OrthoEditSettingsPanel)

        self.__ortho = ortho
        self.__createTools()


    def destroy(self):
        """Must be called when this ``OrthoEditActionToolBar`` is no longer
        needed. Clears references, and calls the base-class ``destroy`` method.
        """
        super().destroy()
        self.__ortho = None


    def __createTools(self):
        """Called when the :attr:`.ViewPanel.profile` property of the
        :class:`.OrthoPanel` changes. Shows/hides edit controls accordingly.
        """

        self.ClearTools(destroy=True, postevent=False)

        ortho      = self.__ortho
        profileObj = ortho.currentProfile

        tools = []
        nav   = []

        for spec in _TOOLBAR_SPECS:

            if spec == 'div':
                tools.append(fsltoolbar.ToolBarDivider(self,
                                                       height=24,
                                                       orient=wx.HORIZONTAL))
                continue

            if spec.key == 'toggleEditPanel': target = self
            else:                             target = profileObj

            widget    = props.buildGUI(self, target, spec)
            navWidget = widget

            if spec.label is not None:
                widget = self.MakeLabelledTool(widget, spec.label)

            tools.append(widget)
            nav  .append(navWidget)

        self.SetTools(tools)
        self.setNavOrder(nav)


_ICONS = {

    'locationFollowsMouse' : [
        fslicons.findImageFile('locationFollowsMouseHighlight24'),
        fslicons.findImageFile('locationFollowsMouse24')],

    'showSelection' : [
        fslicons.findImageFile('showSelectionHighlight24'),
        fslicons.findImageFile('showSelection24')],

    'toggleEditPanel'    : [fslicons.findImageFile('editSpannerHighlight24'),
                            fslicons.findImageFile('editSpanner24')],
    'undo'               : fslicons.findImageFile('undo24'),
    'redo'               : fslicons.findImageFile('redo24'),
    'createMask'         : fslicons.findImageFile('new24'),
    'clearSelection'     : fslicons.findImageFile('clearSelection24'),
    'fillSelection'      : fslicons.findImageFile('fillSelection24'),
    'eraseSelection'     : fslicons.findImageFile('eraseSelection24'),
    'copyPasteData'      : [fslicons.findImageFile('copyDataHighlight24'),
                            fslicons.findImageFile('copyData24')],
    'copyPasteSelection' : [fslicons.findImageFile('copySelectionHighlight24'),
                            fslicons.findImageFile('copySelection24')]

}


_TOOLTIPS = {

    'locationFollowsMouse' : fsltooltips.properties['OrthoEditProfile.'
                                                    'locationFollowsMouse'],
    'showSelection'        : fsltooltips.properties['OrthoEditProfile.'
                                                    'showSelection'],

    'toggleEditPanel'    : fsltooltips.actions['OrthoPanel.'
                                               'toggleEditPanel'],
    'undo'               : fsltooltips.actions['OrthoEditProfile.'
                                               'undo'],
    'redo'               : fsltooltips.actions['OrthoEditProfile.'
                                               'redo'],
    'createMask'         : fsltooltips.actions['OrthoEditProfile.'
                                               'createMask'],
    'clearSelection'     : fsltooltips.actions['OrthoEditProfile.'
                                               'clearSelection'],
    'fillSelection'      : fsltooltips.actions['OrthoEditProfile.'
                                               'fillSelection'],
    'eraseSelection'     : fsltooltips.actions['OrthoEditProfile.'
                                               'eraseSelection'],
    'copyPasteData'      : fsltooltips.actions['OrthoEditProfile.'
                                               'copyPasteData'],
    'copyPasteSelection' : fsltooltips.actions['OrthoEditProfile.'
                                               'copyPasteSelection'],
}

_TOOLBAR_SPECS = [

    actions.ToggleActionButton(
        'toggleEditPanel',
        actionKwargs={'floatPane' : True},
        icon=_ICONS['toggleEditPanel'],
        tooltip=_TOOLTIPS['toggleEditPanel']),
    actions.ActionButton(
        'createMask',
        icon=_ICONS['createMask'],
        tooltip=_TOOLTIPS['createMask']),

    'div',

    props.Widget(
        'locationFollowsMouse',
        icon=_ICONS['locationFollowsMouse'],
        tooltip=_TOOLTIPS['locationFollowsMouse']),

    'div',


    actions.ActionButton(
        'undo',
        icon=_ICONS['undo'],
        tooltip=_TOOLTIPS['undo']),
    actions.ActionButton(
        'redo',
        icon=_ICONS['redo'],
        tooltip=_TOOLTIPS['redo']),

    'div',

    props.Widget(
        'showSelection',
        icon=_ICONS['showSelection'],
        tooltip=_TOOLTIPS['showSelection'],
        dependencies=['drawMode'],
        enabledWhen=lambda i, m: not m),
    actions.ActionButton(
        'clearSelection',
        icon=_ICONS['clearSelection'],
        tooltip=_TOOLTIPS['clearSelection']),
    actions.ActionButton(
        'fillSelection',
        icon=_ICONS['fillSelection'],
        tooltip=_TOOLTIPS['fillSelection']),
    actions.ActionButton(
        'eraseSelection',
        icon=_ICONS['eraseSelection'],
        tooltip=_TOOLTIPS['eraseSelection']),

    'div',

    actions.ToggleActionButton(
        'copyPasteData',
        icon=_ICONS['copyPasteData'],
        tooltip=_TOOLTIPS['copyPasteData']),

    actions.ToggleActionButton(
        'copyPasteSelection',
        icon=_ICONS['copyPasteSelection'],
        tooltip=_TOOLTIPS['copyPasteSelection']),
]



#!/usr/bin/env python
#
# ortholabels.py - The OrthoLabels class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoLabels` class, which manages
anatomical orientation labels for an :class:`.OrthoPanel`.

This logic is independent from the :class:`.OrthoPanel` so it can be used in
off-screen rendering (see :mod:`.render`).
"""


import fsl.data.constants as constants


class OrthoLabels:
    """The ``OrthoLabels`` class manages anatomical orientation labels which
    are displayed on a set of three :class:`.SliceCanvas` instances, one for
    each plane in the display coordinate system, typically within an
    :class:`.OrthoPanel`.

    The ``OrthoLabels`` class uses :class:`.annotations.Text` annotations,
    showing the user the anatomical orientation of the display on each
    canvas. These labels are only shown if the currently selected overlay (as
    dicated by the :attr:`.DisplayContext.selectedOverlay` property) is a
    :class:`.Image` instance, **or** the
    :meth:`.DisplayOpts.referenceImage` property for the currently selected
    overlay returns an :class:`.Image` instance.
    """


    def __init__(self,
                 overlayList,
                 displayCtx,
                 orthoOpts,
                 *canvases):
        """Create an ``OrthoLabels`` instance.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        :arg orthoOpts:   :class:`.OrthoOpts` instance which contains
                          display settings.
        :arg canvases:    The :class:`.SliceCanvas` instances to be labelled.
        """

        self.__name        = '{}_{}'.format(type(self).__name__, id(self))
        self.__orthoOpts   = orthoOpts
        self.__displayCtx  = displayCtx
        self.__overlayList = overlayList

        # labels is a list of dicts, one
        # for each canvas, containing Text
        # annotations to show anatomical
        # orientation
        annots = [{} for c in canvases]

        self.__canvases = canvases
        self.__annots   = annots

        # Create the Text annotations
        for side in ('left', 'right', 'top', 'bottom'):
            for canvas, cannots in zip(canvases, annots):
                annot         = canvas.getAnnotations()
                cannots[side] = annot.text('', 0, 0, hold=True)

        # Initialise the display properties
        # of each Text annotation
        for cannots in annots:
            cannots['left']  .halign = 'left'
            cannots['right'] .halign = 'right'
            cannots['top']   .halign = 'centre'
            cannots['bottom'].halign = 'centre'

            cannots['left']  .valign = 'centre'
            cannots['right'] .valign = 'centre'
            cannots['top']   .valign = 'top'
            cannots['bottom'].valign = 'bottom'

            cannots['left']  .x = 0.0
            cannots['left']  .y = 0.5
            cannots['right'] .x = 1.0
            cannots['right'] .y = 0.5
            cannots['bottom'].x = 0.5
            cannots['bottom'].y = 0.0
            cannots['top']   .x = 0.5
            cannots['top']   .y = 1.0

            # Keep cannots 5 pixels away
            # from the canvas edges
            cannots['left']  .off = ( 5,  0)
            cannots['right'] .off = (-5,  0)
            cannots['top']   .off = ( 0, -5)
            cannots['bottom'].off = ( 0,  5)

        # Add listeners to properties
        # that need to trigger a label
        # refresh.
        name = self.__name

        # Make immediate so the label
        # annotations get updated before
        # a panel refresh occurs (where
        # the latter is managed by the
        # OrthoPanel).
        refreshArgs = {
            'name'      : name,
            'callback'  : self.__refreshLabels,
            'immediate' : True
        }

        for c in canvases:
            c.opts.addListener('invertX', **refreshArgs)
            c.opts.addListener('invertY', **refreshArgs)

        orthoOpts  .addListener('showLabels',       **refreshArgs)
        orthoOpts  .addListener('labelSize',        **refreshArgs)
        orthoOpts  .addListener('fgColour',         **refreshArgs)
        displayCtx .addListener('selectedOverlay',  **refreshArgs)
        displayCtx .addListener('displaySpace',     **refreshArgs)
        displayCtx .addListener('radioOrientation', **refreshArgs)
        overlayList.addListener('overlays', name, self.__overlayListChanged)


    def destroy(self):
        """Must be called when this ``OrthoLabels`` instance is no longer
        needed.
        """

        name        = self.__name
        overlayList = self.__overlayList
        displayCtx  = self.__displayCtx
        orthoOpts   = self.__orthoOpts
        canvases    = self.__canvases
        annots      = self.__annots

        self.__overlayList = None
        self.__displayCtx  = None
        self.__orthoOpts   = None
        self.__canvases    = None
        self.__annots      = None

        orthoOpts  .removeListener('showLabels',       name)
        orthoOpts  .removeListener('labelSize',        name)
        orthoOpts  .removeListener('fgColour',         name)
        displayCtx .removeListener('selectedOverlay',  name)
        displayCtx .removeListener('displaySpace',     name)
        displayCtx .removeListener('radioOrientation', name)
        overlayList.removeListener('overlays',         name)

        for c in canvases:
            c.opts.removeListener('invertX', name)
            c.opts.removeListener('invertY', name)

        # The _overlayListChanged method adds
        # listeners to individual overlays,
        # so we have to remove them too
        for ovl in overlayList:
            opts = displayCtx.getOpts(ovl)
            opts.removeListener('bounds', name)

        # Destroy the Text annotations
        for a in annots:
            for text in a.values():
                text.destroy()


    def refreshLabels(self):
        """Forces the label annotations to be refreshed."""
        self.__refreshLabels()


    def __overlayListChanged(self, *a):
        """Called when the :class:`.OverlayList` changes. Registers a listener
        on the attr:.DisplayOpts.bounds` property of every overlay in the list,
        so the labels are refreshed when any overlay bounds change.
        """

        for ovl in self.__overlayList:

            opts = self.__displayCtx.getOpts(ovl)

            # Update anatomy labels when
            # overlay bounds change
            opts.addListener('bounds',
                             self.__name,
                             self.__refreshLabels,
                             overwrite=True)

        # When the list becomes empty, or
        # an overlay is added to an empty
        # list, the DisplayContext.selectedOverlay
        # will not change, and __refreshLabels
        # will thus not get called. So we call
        # it here.
        if len(self.__overlayList) in (0, 1):
            self.__refreshLabels()


    def __refreshLabels(self, *a):
        """Updates the attributes of the :class:`.Text` anatomical orientation
        annotations on each :class:`.SliceCanvas`.
        """

        displayCtx = self.__displayCtx
        sopts      = self.__orthoOpts
        overlay    = displayCtx.getSelectedOverlay()

        canvases = self.__canvases
        annots   = self.__annots

        for cannots in annots:
            for text in cannots.values():
                text.enabled = sopts.showLabels and (overlay is not None)

        if not sopts.showLabels or overlay is None:
            return

        opts = displayCtx.getOpts(overlay)

        # Calculate all of the xyz
        # labels for this overlay
        labels, orients              = opts.getLabels()
        xlo, ylo, zlo, xhi, yhi, zhi = labels
        vertOrient                   = len(xlo) > 1

        fontSize = sopts.labelSize
        fgColour = tuple(sopts.fgColour)

        # If any axis orientation is unknown, make
        # the foreground colour red, to highlight
        # the unknown orientation.
        if constants.ORIENT_UNKNOWN in orients:
            fgColour = (1, 0, 0, 1)

        # A list, with one entry for each canvas,
        # and with each entry of the form:
        #
        #   [[xlo, xhi], [ylo, yhi]]
        #
        # containing the low/high labels for the
        # horizontal (x) and vertical (y) canvas
        # axes.
        canvasLabels = []
        for canvas in canvases:

            cax = 'xyz'[canvas.opts.zax]

            if   cax == 'x': clabels = [[ylo, yhi], [zlo, zhi]]
            elif cax == 'y': clabels = [[xlo, xhi], [zlo, zhi]]
            elif cax == 'z': clabels = [[xlo, xhi], [ylo, yhi]]

            if canvas.opts.invertX: clabels[0] = [clabels[0][1], clabels[0][0]]
            if canvas.opts.invertY: clabels[1] = [clabels[1][1], clabels[1][0]]

            canvasLabels.append(clabels)

        # Update the text annotation properties
        for canvas, cannots, clabels in zip(canvases, annots, canvasLabels):

            cax = 'xyz'[canvas.opts.zax]

            cannots['left']  .text = clabels[0][0]
            cannots['right'] .text = clabels[0][1]
            cannots['bottom'].text = clabels[1][0]
            cannots['top']   .text = clabels[1][1]

            if   cax == 'x': show = sopts.showXCanvas
            elif cax == 'y': show = sopts.showYCanvas
            elif cax == 'z': show = sopts.showZCanvas

            for side in ['left', 'right', 'bottom', 'top']:

                cannots[side].enabled  = show
                cannots[side].fontSize = fontSize
                cannots[side].colour   = fgColour

            if vertOrient:
                cannots['left'] .angle = 90
                cannots['right'].angle = 90
